/*--------------------------------------------------------------------
 * $Id: ck.c,v 1.7 2001/09/25 16:58:20 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../src_c/ascii.h"
#include "../src_c/fortran_and_c.h"
#include "../src_f/f77-uscore.h"
#include "../src_f/avhrr_kratz.h"
#include "uvspec.h"
#include "ckdfu.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


/* prototypes of internal functions */
static int x_section (int ispcs, int ipress, int itemp, 
		      double presslog, double temp, int iband, char *path, 
		      int *ngauss, double *xabs, double *weight);

static int x_section_h2o (int *ipress, int *itemp, int *irho,
			  double *presslog, double *temp, 
			  double *rholog, int *iband, char *path, 
			  int *ngauss, double *xabs, double *weight);

static int hunt(float *xx, int *n, double *x, int *jlo);

static int kato_crs (double press, double temp, double dens, double dz,
		     ck_struct *crs, int iband, int ispcs, char *path,
		     double *extinction, double *weight, int *ngauss);

static int fu_profile (float *temper, float *press, float *z, float **dens, int nlev,
		       int h2o_cont, float umco2, float umch4, float umn2o,
		       float umf11, float umf12, float umf22,
		       ck_profile *ck);


static int avhrr_kratz (int channel, int interval,
			float *z0, float *p0, float *t0, float *u0, float *ux, 
			int nlev, float co2, float n2o, float f11, float f12, float ch4, 
			double **od, double *wght, int *ntau);
     

static int avhrr_kratz_profile (float *temper, float *press, float *z, float **dens, int nlev, 
				float umco2, float umch4, float umn2o, float umf11, float umf12,
				ck_profile *ck);

static int generic_profile (char *filename, float *z, int nlev, int *nwght, ck_profile *ck);

#define M_AIR 28.977
#define M_H2O 18.015
#define M_O3  48.000

#define AMU 1.660438E-27

#define N_MOL_CKDFU 6.0235e23



/*****************************************************************/
/* Read the Kato et al. wavelength grid and cross section tables */
/* and write the data to ck_struct *ck                           */
/*****************************************************************/

int kato_readtables (ck_struct *ck, char *path, char *photon_filename)
{
  int i=0, iband=0, status=0;
  int iv=0, iq1=0, iq2=0, iq3=0, iq4=0;
  int nq1=0, nq2=0, nq3=0, nq4=0;
  char filename[FILENAME_MAX]="";

  int n_tmp=0;
  double *tmp1=NULL, *tmp2=NULL, *tmp3=NULL, *tmp4=NULL, *tmp5=NULL, *tmp6=NULL;


  ck->scheme = CK_KATO;

  /* read wavelength  information */

  strcpy(filename, path);
  strcat(filename, "correlated_k/kato/wvl.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &(ck->n_wvl));
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* allocate memory for wavelengths */
  ck->wvlc = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl1 = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl2 = calloc(ck->n_wvl+1, sizeof(double));


  /* 4 components (1,2,3,4) in Kato et al. [1999] */
  ck->comp = calloc (KATO_COMPONENTS+1, sizeof (double *));
  ck->wght = calloc (KATO_COMPONENTS+1, sizeof (int *));

  /* check structure (bands need to be sorted in ascending order  */
  /* and are assumed to start with 1) and copy data to wavelength */
  /* arrays.                                                      */

  for (i=0; i<ck->n_wvl; i++) {
    iband = (int) (tmp1[i]+0.5); 
    
    if (iband != i+1) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n", filename, iband);
      return -1;
    }
    
    /* copy and convert to nm */
    (ck->wvl1)[iband] = tmp2[i]*1000.0;
    (ck->wvl2)[iband] = tmp3[i]*1000.0;
    (ck->wvlc)[iband] = ((ck->wvl1)[iband] + (ck->wvl2)[iband]) / 2.0;
  }

  free(tmp1); free(tmp2); free(tmp3);

  
  /* allocate memory for cross sections */
  ck->comp[KATO_CO2] = calloc (ck->n_wvl+1, sizeof(double));
  ck->comp[KATO_O3]  = calloc (ck->n_wvl+1, sizeof(double));
  ck->comp[KATO_O2]  = calloc (ck->n_wvl+1, sizeof(double));
  ck->comp[KATO_H2O] = calloc (ck->n_wvl+1, sizeof(double));
  
  ck->wght[KATO_CO2] = calloc (ck->n_wvl+1, sizeof(int));
  ck->wght[KATO_O3]  = calloc (ck->n_wvl+1, sizeof(int));
  ck->wght[KATO_O2]  = calloc (ck->n_wvl+1, sizeof(int));
  ck->wght[KATO_H2O] = calloc (ck->n_wvl+1, sizeof(int));
  
    
  /* read CO2 data */

  strcpy(filename, path);
  strcat(filename, "correlated_k/kato/co2.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    (ck->comp)[KATO_CO2][iband] = tmp3[i];
    (ck->wght)[KATO_CO2][iband] = (int) (tmp2[i]+0.5);
  }

  free(tmp1); free(tmp2); free(tmp3);



  /* read O3 data */
  strcpy(filename, path);
  strcat(filename, "correlated_k/kato/o3.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    (ck->comp)[KATO_O3][iband] = tmp3[i];
    (ck->wght)[KATO_O3][iband] = (int) (tmp2[i]+0.5);
  }

  free(tmp1); free(tmp2); free(tmp3);


  /* read O2 data */
  strcpy(filename, path);
  strcat(filename, "correlated_k/kato/o2.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    ck->comp[KATO_O2][iband] = tmp3[i];
    ck->wght[KATO_O2][iband] = (int) (tmp2[i]+0.5);
  }

  free(tmp1); free(tmp2); free(tmp3);


  /* read H2O data */
  strcpy(filename, path);
  strcat(filename, "correlated_k/kato/h2o.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    (ck->comp)[KATO_H2O][iband] = tmp3[i];
    (ck->wght)[KATO_H2O][iband] = (int) (tmp2[i]+0.5);
  }

  free(tmp1); free(tmp2); free(tmp3);


  /* read photon fractions */

  if (strlen(photon_filename)==0) {
    /* default photon file */
    strcpy(photon_filename, path);
    strcat(photon_filename, "correlated_k/kato/photons.dat");
  }

  status = read_6c_file (photon_filename, &tmp1, &tmp2, &tmp3, &tmp4, &tmp5, &tmp6, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, photon_filename);
    return status;
  }

  /* allocate memory for x_solar[iv][iq1][iq2][iq3][iq4] */

  ck->x_solar = calloc (ck->n_wvl+1, sizeof (double ****));
  
  for (iv=1; iv<=ck->n_wvl; iv++) {
    
    nq1 = (ck->wght)[KATO_CO2][iv];
    nq2 = (ck->wght)[KATO_O3] [iv];
    nq3 = (ck->wght)[KATO_O2] [iv];
    nq4 = (ck->wght)[KATO_H2O][iv];

    ((double *****) ck->x_solar)[iv] = calloc (nq1+1, sizeof(double ***));
    
    for (iq1=1; iq1<=nq1; iq1++) {
      ((double *****) ck->x_solar)[iv][iq1] = calloc (nq2+1, sizeof(double **));

      for (iq2=1; iq2<=nq2; iq2++) {
	((double *****) ck->x_solar)[iv][iq1][iq2] = calloc (nq3+1, sizeof(double *));
	
	for (iq3=1; iq3<=nq3; iq3++)
	  ((double *****) ck->x_solar)[iv][iq1][iq2][iq3] = calloc (nq4+1, sizeof(double));
      }
    }
  }

  for (i=0; i<n_tmp; i++) {
    iv  = (int) (tmp1[i]+0.5); 
    iq1 = (int) (tmp2[i]+0.5); 
    iq2 = (int) (tmp3[i]+0.5); 
    iq3 = (int) (tmp4[i]+0.5); 
    iq4 = (int) (tmp5[i]+0.5); 

    ((double *****) ck->x_solar)[iv][iq1][iq2][iq3][iq4] = tmp6[i];
  }


  free(tmp1); free(tmp2); free(tmp3);
  free(tmp4); free(tmp5); free(tmp6);


  return 0;
}



/*****************************************************************/
/* Read the Fu et al. wavelength grid and cross section tables   */
/* and write the data to ck_struct *ck                           */
/*****************************************************************/

int fu_readtables (ck_struct *ck, char *path, char *photon_filename)
{
  int i=0, iband=0, iwght=0, status=0;
  int n_tmp=0;
  double *tmp1=NULL, *tmp2=NULL, *tmp3=NULL, *tmp4=NULL;
  char filename[FILENAME_MAX]="";

  char solar_filename  [FILENAME_MAX]="";
  char thermal_filename[FILENAME_MAX]="";


  ck->scheme = CK_FU;

  /* read wavelength  information */

  strcpy(filename, path);
  strcat(filename, "correlated_k/fu/wvl.dat");

  status = read_4c_file (filename, &tmp1, &tmp2, &tmp3, &tmp4, &(ck->n_wvl));
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* allocate memory for wavelengths */
  ck->wvlc = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl1 = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl2 = calloc(ck->n_wvl+1, sizeof(double));

  /* Only one component is considered explicitely */
  ck->comp = calloc (FU_COMPONENTS+1, sizeof (double *));
  ck->wght = calloc (FU_COMPONENTS+1, sizeof (int *));

  /* check structure (bands need to be sorted in ascending order  */
  /* and are assumed to start with 1) and copy data to wavelength */
  /* arrays.                                                      */

  for (i=0; i<ck->n_wvl; i++) {
    iband = (int) (tmp1[i]+0.5); 
    
    if (iband != i+1) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n", filename, iband);
      return -1;
    }
    
    /* copy and convert to nm */
    (ck->wvlc)[iband] = tmp2[i]*1000.0;
    (ck->wvl1)[iband] = tmp3[i]*1000.0;
    (ck->wvl2)[iband] = tmp4[i]*1000.0;
  }


  
  /* allocate memory for cross sections */
  ck->comp[FU_WGHT] = calloc (ck->n_wvl+1, sizeof(double));
  ck->wght[FU_WGHT] = calloc (ck->n_wvl+1, sizeof(int));

  /* free memory */
  free(tmp1); free(tmp2); free(tmp3); free(tmp4);
  
    
  /* read weights */
  strcpy(filename, path);
  strcat(filename, "correlated_k/fu/quad.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    (ck->comp)[FU_WGHT][iband] = tmp3[i];
    (ck->wght)[FU_WGHT][iband] = (int) (tmp2[i]+0.5);
  }

  /* free memory */
  free(tmp1); free(tmp2); free(tmp3);


  /* read photon fractions */
  if (strlen(photon_filename)==0) {
    /* default photon files */

    strcpy(solar_filename, path);
    strcat(solar_filename, "correlated_k/fu/x_solar.dat");

    strcpy(thermal_filename, path);
    strcat(thermal_filename, "correlated_k/fu/x_thermal.dat");
  }
  else {
    strcpy(solar_filename, photon_filename);
    strcpy(thermal_filename, photon_filename);
  }


  /* solar */
  status = read_3c_file (solar_filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, solar_filename);
    return status;
  }

  /* allocate memory */
  ck->x_solar = calloc (ck->n_wvl+1, sizeof (double *));
  for (i=0; i<=ck->n_wvl; i++)
    ((double **) ck->x_solar)[i] = calloc (FU_MAXINT+1, sizeof(double));

  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    iwght = (int) (tmp2[i]+0.5); 
    
    ((double **) ck->x_solar)[iband][iwght] = tmp3[i];
  }


  /* thermal */
  status = read_3c_file (thermal_filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, thermal_filename);
    return status;
  }

  /* allocate memory */
  ck->x_thermal = calloc (ck->n_wvl+1, sizeof (double *));
  for (i=0; i<=ck->n_wvl; i++)
    ((double **) ck->x_thermal)[i] = calloc (FU_MAXINT+1, sizeof(double));

  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    iwght = (int) (tmp2[i]+0.5); 
    
    ((double **) ck->x_thermal)[iband][iwght] = tmp3[i];
  }

  return 0;
}



/*****************************************************************/
/* Read the Kratz AVHRR wavelength grid and cross section tables */
/* and write the data to ck_struct *ck                           */
/*****************************************************************/

int avhrr_kratz_readtables (ck_struct *ck, char *path, char *photon_filename)
{
  int i=0, iband=0, iwght=0, status=0;
  char filename[FILENAME_MAX]="";
  int n_tmp=0;
  double *tmp1=NULL, *tmp2=NULL, *tmp3=NULL;

  char solar_filename  [FILENAME_MAX]="";
  char thermal_filename[FILENAME_MAX]="";


  ck->scheme = CK_AVHRR_KRATZ;

  /* read wavelength  information */

  strcpy(filename, path);
  strcat(filename, "correlated_k/kratz/avhrr_wvl.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &(ck->n_wvl));
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* allocate memory for wavelengths */
  ck->wvlc = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl1 = calloc(ck->n_wvl+1, sizeof(double));
  ck->wvl2 = calloc(ck->n_wvl+1, sizeof(double));


  /* Only one component is considered explicitely */
  ck->comp = calloc (AVHRR_KRATZ_COMPONENTS+1, sizeof (double *));
  ck->wght = calloc (AVHRR_KRATZ_COMPONENTS+1, sizeof (int *));

  /* check structure (bands need to be sorted in ascending order  */
  /* and are assumed to start with 1) and copy data to wavelength */
  /* arrays.                                                      */

  for (i=0; i<ck->n_wvl; i++) {
    iband = (int) (tmp1[i]+0.5); 
    
    if (iband != i+1) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n", filename, iband);
      return -1;
    }
    
    /* copy and convert to nm */
    (ck->wvl1)[iband] = tmp2[i]*1000.0;
    (ck->wvl2)[iband] = tmp3[i]*1000.0;
    (ck->wvlc)[iband] = ((ck->wvl1)[iband] + (ck->wvl2)[iband]) / 2.0;
  }

  free(tmp1); free(tmp2); free(tmp3);

  
  /* allocate memory for cross sections */
  ck->comp[AVHRR_KRATZ_WGHT] = calloc (ck->n_wvl+1, sizeof(double));
  ck->wght[AVHRR_KRATZ_WGHT] = calloc (ck->n_wvl+1, sizeof(int));
  
    
  /* read weights */
  strcpy(filename, path);
  strcat(filename, "correlated_k/kratz/avhrr_quad.dat");

  status = read_3c_file (filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    if (iband > ck->n_wvl) {
      fprintf (stderr, " ... inconsistency in %s, band %d\n",
	       filename, iband);
      return -1;
    }
    
    (ck->comp)[FU_WGHT][iband] = tmp3[i];
    (ck->wght)[FU_WGHT][iband] = (int) (tmp2[i]+0.5);
  }

  /* free memory */
  free(tmp1); free(tmp2); free(tmp3);


  /* read photon fractions */
  if (strlen(photon_filename)==0) {
    /* default photon files */

    strcpy(solar_filename, path);
    strcat(solar_filename, "correlated_k/kratz/x_solar.dat");

    strcpy(thermal_filename, path);
    strcat(thermal_filename, "correlated_k/kratz/x_thermal.dat");
  }
  else {
    strcpy(solar_filename, photon_filename);
    strcpy(thermal_filename, photon_filename);
  }


  /* solar */
  status = read_3c_file (solar_filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, solar_filename);
    return status;
  }

  /* allocate memory */
  ck->x_solar = calloc (ck->n_wvl+1, sizeof (double *));
  for (i=0; i<=ck->n_wvl; i++)
    ((double **) ck->x_solar)[i] = calloc (AVHRR_KRATZ_MAXINT+1, sizeof(double));

  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    iwght = (int) (tmp2[i]+0.5); 
    
    ((double **) ck->x_solar)[iband][iwght] = tmp3[i];
  }


  /* thermal */
  status = read_3c_file (thermal_filename, &tmp1, &tmp2, &tmp3, &n_tmp);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, thermal_filename);
    return status;
  }

  /* allocate memory */
  ck->x_thermal = calloc (ck->n_wvl+1, sizeof (double *));
  for (i=0; i<=ck->n_wvl; i++)
    ((double **) ck->x_thermal)[i] = calloc (AVHRR_KRATZ_MAXINT+1, sizeof(double));

  for (i=0; i<n_tmp; i++) {
    iband = (int) (tmp1[i]+0.5); 
    iwght = (int) (tmp2[i]+0.5); 
    
    ((double **) ck->x_thermal)[iband][iwght] = tmp3[i];
  }

  return 0;
}



/*****************************************************************/
/* Read a generic CK table and write the data to ck_struct *ck   */
/*****************************************************************/

int ck_generic_readtables (ck_struct *ck, char *filename, char *photon_filename)
{
  int iv=0, status=0;

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  int ncid=0;
  int idd_nwvn=0, idd_nlyr=0, idd_maxc=0;
  int id_wvnlo=0, id_wvnhi=0, id_nc=0;

  int start[1] = {0};
  int count[1] = {0};

  int nlyr=0, maxc=0;

  double *wvnlo=NULL, *wvnhi=NULL;
  int *nc=NULL;

  ck->scheme = CK_GENERIC;

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }


  /* get dimension id for "nwvn" */
  status = nc_inq_dimid (ncid, "nwvn", &idd_nwvn);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nwvn" */
  status = nc_inq_dimlen (ncid, idd_nwvn, &(ck->n_wvl));
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }



  /* get dimension id for "nlyr" */
  status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlyr" */
  status = nc_inq_dimlen (ncid, idd_nlyr, &nlyr);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }



  /* get dimension id for "maxc" */
  status = nc_inq_dimid (ncid, "maxc", &idd_maxc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "maxc" */
  status = nc_inq_dimlen (ncid, idd_maxc, &maxc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  count[0] = ck->n_wvl;
  
  /* allocate memory for fields */
  wvnlo = calloc (ck->n_wvl, sizeof(double));
  wvnhi = calloc (ck->n_wvl, sizeof(double));
  nc    = calloc (ck->n_wvl, sizeof(int));

  ck->wvl1 = calloc (ck->n_wvl+1, sizeof(double));
  ck->wvl2 = calloc (ck->n_wvl+1, sizeof(double));
  ck->wvlc = calloc (ck->n_wvl+1, sizeof(double));
  

  /* get variable id for "wvnlo" */
  status = nc_inq_varid (ncid, "wvnlo", &id_wvnlo);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* read "wvnlo" */
  status = nc_get_vara_double (ncid, id_wvnlo, start, count, wvnlo);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
  }

  /* get variable id for "wvnhi" */
  status = nc_inq_varid (ncid, "wvnhi", &id_wvnhi);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* read "wvnhi" */
  status = nc_get_vara_double (ncid, id_wvnhi, start, count, wvnhi);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
  }


  for (iv=1; iv<=ck->n_wvl; iv++) {
    ck->wvlc[iv] = 1.0E7 / (0.5 * (wvnlo[iv-1] + wvnhi[iv-1]));
    ck->wvl1[iv] = 1.0E7 / wvnhi[iv-1];
    ck->wvl2[iv] = 1.0E7 / wvnlo[iv-1];
    
  }


  ck->comp = calloc (GENERIC_COMPONENTS+1, sizeof(double *));
  ck->wght = calloc (GENERIC_COMPONENTS+1, sizeof(int *));

  ck->comp[GENERIC_WGHT] = calloc (ck->n_wvl+1, sizeof(double));
  ck->wght[GENERIC_WGHT] = calloc (ck->n_wvl+1, sizeof(int));


  /* get variable id for "nc" */
  status = nc_inq_varid (ncid, "nc", &id_nc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* read "nc" */
  status = nc_get_vara_int (ncid, id_nc, start, count, nc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
  }

  for (iv=1; iv<=ck->n_wvl; iv++) {
    ck->comp[GENERIC_WGHT][iv] = -1.0;
    ck->wght[GENERIC_WGHT][iv] = nc[iv-1];
  }

  free (wvnlo);
  free (wvnhi);
  free (nc);

  nc_close(ncid);

  /* allocate memory for photon fractions */
  /* ??? need to read data from file ???  */
  ck->x_solar   = calloc (ck->n_wvl+1, sizeof (double *));
  ck->x_thermal = calloc (ck->n_wvl+1, sizeof (double *));

  for (iv=0; iv<=ck->n_wvl; iv++) {
    ((double **) ck->x_solar)[iv]   = calloc (ck->wght[GENERIC_WGHT][iv], sizeof(double));
    ((double **) ck->x_thermal)[iv] = calloc (ck->wght[GENERIC_WGHT][iv], sizeof(double));
  }



#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif
  
  return 0;  /* ok */
}




int sbdart_readtables (ck_struct *ck, int nwvl, char *path, char *photon_filename)
{
  int iv=0, iq=0;

  ck->scheme = CK_SBDART;
  ck->n_wvl = nwvl;

  fprintf (stderr, " ... sbdart_readtables\n");
  fprintf (stderr, " ... using %d wavelengths\n", nwvl);


  /* allocate memory for photon numbers and set to 1 */
  /* ???  need to invent something better here   ??? */

  ck->x_solar = calloc (ck->n_wvl+1, sizeof (double *));
  for (iv=0; iv<=ck->n_wvl; iv++) {
    ((double **) ck->x_solar)[iv] = calloc (SBDART_MAXINT+1, sizeof(double));
   
    for (iq=0; iq<=SBDART_MAXINT; iq++)
      ((double **) ck->x_solar)[iv][iq] = 1;
  }

  ck->x_thermal = calloc (ck->n_wvl+1, sizeof (double *));
  for (iv=0; iv<=ck->n_wvl; iv++) {
    ((double **) ck->x_thermal)[iv] = calloc (SBDART_MAXINT+1, sizeof(double));
   
    for (iq=0; iq<=SBDART_MAXINT; iq++)
      ((double **) ck->x_thermal)[iv][iq] = 1;
  }

  return 0;
}
  
  
/*****************************************************************/
/* Calculate cross section for a given temperature, pressure,    */
/* and concentration according to Kato et al. [1999]             */
/*****************************************************************/

static int kato_crs (double press, double temp, double dens, double dz,
		     ck_struct *crs, int iband, int ispcs, char *path,
		     double *extinction, double *weight, int *ngauss)
{
  static float xgrid[17] = { 3.02119, 2.92117, 2.82086, 2.72099, 
			     2.62118, 2.52114, 2.42160, 2.22531, 
			     2.02531, 1.82543, 1.62531, 1.42488, 
			     1.22531, 1.02531, 0.00000, -1.0000,
			     -2.0000 };

  static float ygrid[11] = { 120.0, 140.0, 160.0, 180.0, 
			     200.0, 220.0, 240.0, 260.0, 
			     280.0, 300.0, 320.0 };
  
  static float zgrid[11] = { 24.1761, 23.6721, 23.1761,
			     22.6721, 22.1761, 21.6721, 
			     21.1761, 20.6721, 20.1761,
			     19.6721, 19.1761 };
  
  
  /* Local variables */
  int irho=0, itemp=0, ipress=0, tmp=0;
  double rholog=0, presslog=0;
  
  

  if (crs->comp[ispcs][iband] < 0) {   /* if correlated k in this band */

    presslog = log10(press);

    tmp=17;
    hunt(xgrid, &tmp, &presslog, &ipress);

    tmp=11;
    hunt(ygrid, &tmp, &temp, &itemp);
    
    *ngauss = crs->wght[ispcs][iband];
    
    if (ispcs == KATO_H2O) {
      rholog = log10(dens);
      
      tmp=11;
      hunt(zgrid, &tmp, &rholog, &irho);
      
      
      if (iband == 10)
	x_section (ispcs, ipress, itemp, presslog, temp, iband, 
		   path, ngauss, extinction, weight);
      else
	x_section_h2o (&ipress, &itemp, &irho, &presslog, &temp, &
		       rholog, &iband, path, ngauss, extinction, weight);
    } 
    else {
      x_section (ispcs, ipress, itemp, presslog, temp, iband, 
		 path, ngauss, extinction, weight);
    }
  }
  else {
    *ngauss = 1;
    weight[0]     = crs->wght[ispcs][iband];
    extinction[0] = crs->comp[ispcs][iband];
  }

  return 0;
}




/*****************************************************************/
/* Calculate cross section for a given temperature, pressure,    */
/* and concentration                                             */
/*****************************************************************/

static int fu_profile (float *temper, float *press, float *z, float **dens, int nlev, 
		       int h2o_cont, float umco2, float umch4, float umn2o,
		       float umf11, float umf12, float umf22,
		       ck_profile *ck)
{

  float *rhoair=NULL, *rhoo3=NULL, *rhoh2o=NULL;
  float *pres=NULL, *temp=NULL;

  int lc=0, status=0;


  /* allocate memory for density profiles */
  rhoair = calloc (nlev, sizeof(float));
  rhoh2o = calloc (nlev, sizeof(float));
  rhoo3  = calloc (nlev, sizeof(float));
  temp   = calloc (nlev, sizeof(float));
  pres   = calloc (nlev, sizeof(float));


  for (lc=0;lc<nlev;lc++) {
    rhoair[lc] = dens[MOL_AIR][lc] * M_AIR / N_MOL_CKDFU * 1.0e6;
    rhoh2o[lc] = dens[MOL_H2O][lc] * M_H2O / N_MOL_CKDFU * 1.0e6;
    rhoo3 [lc] = dens[MOL_O3] [lc] * M_O3  / N_MOL_CKDFU * 1.0e6;
    temp  [lc] = temper[lc];
    pres[lc]   = press[lc];
  }

  /* call Fu and Liou routines */

#if HAVE_FULIOU
  status = ckdfu (umco2, umch4, umn2o,
		  umf11, umf12, umf22,
		  rhoair, rhoh2o, rhoo3, pres, temp, z, nlev, h2o_cont,
		  ck);
#else
  fprintf (stderr, "Error, Fu and Liou not supported!\n");
  return -1;
#endif

  if (status<0) {
    fprintf (stderr, "Error %d returned by ckdfu()\n", status);
    return status;
  }
 
  free (rhoair);
  free (rhoh2o);
  free (rhoo3);
  free (temp);
  free (pres);

  return 0; /* ok */
}



/*****************************************************************/
/* C wrapper for David Kratz' AVHRR parameterization; should be  */
/* moved to a separate source file.                              */
/*****************************************************************/

static int avhrr_kratz (int channel, int interval,
			float *z0, float *p0, float *t0, float *u0, float *ux, 
			int nlev, float co2, float n2o, 
			float f11, float f12, float ch4, 
			double **od, double *wght, int *ntau)
{
  int i=0, j=0;

  /* allocate memory for output */
  float *tmp_od   = calloc (KRATZ_MAXNTAU*(nlev-1), sizeof(float));
  float *tmp_wght = calloc (KRATZ_MAXNTAU, sizeof(float));
  float **tmp_od2 = NULL;


#if HAVE_AVHRR
  switch (channel) {
  case 1:
    switch(interval) {
    case 1:
      avhrr11_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 2:
      avhrr12_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 3:
      avhrr13_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 4:
      avhrr14_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 5:
      avhrr15_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    default:
      fprintf (stderr, "error, AVHRR channel %d has no interval %d\n", 
	       channel, interval);
      return -1;
    }
    break;
    
  case 2:
    switch(interval) {
    case 1:
      avhrr21_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 2:
      avhrr22_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 3:
      avhrr23_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 4:
      avhrr24_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    default:
      fprintf (stderr, "error, AVHRR channel %d has no interval %d\n", 
	       channel, interval);
      return -1;
    }
    break;
    
  case 3:
    switch(interval) {
    case 1:
      avhrr31_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 2:
      avhrr32_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 3:
      avhrr33_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 4:
      avhrr34_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    case 5:
      avhrr35_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    default:
      fprintf (stderr, "error, AVHRR channel %d has no interval %d\n", 
	       channel, interval);
      return -1;
    }
    break;
    
  case 4:
    switch(interval) {
    case 1:
      avhrr41_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    default:
      fprintf (stderr, "error, AVHRR channel %d has no interval %d\n", 
	       channel, interval);
      return -1;
    }
    break;

  case 5:
    switch(interval) {
    case 1:
      avhrr51_(z0, p0, t0, u0, ux, &nlev, &co2, &n2o, &f11, &f12, &ch4, 
	       tmp_od, tmp_wght, ntau);
      break;
    default:
      fprintf (stderr, "error, AVHRR channel %d has no interval %d\n", 
	       channel, interval);
      return -1;
    }
    break;
  default:
    fprintf (stderr, "error, unkown AVHRR channel %d\n", channel);
    return -1;
  }
#else
  fprintf (stderr, "Error, AVHRR parameterization not supported!\n");
  return -1;
#endif

  /* convert fortran array to a 2D C array */
  /* tmp_od2 [0 ... nlyr-1][0 ... ntau-1]  */

  tmp_od2 = fortran2c_2D_float_ary(nlev-1, *ntau, tmp_od);
  
  /* copy arrays to their final destination */
    
  for (j=0; j<*ntau; j++) {
    wght[j] = tmp_wght[j];
    for (i=0; i<nlev-1; i++)   /* flip profile */
      od[nlev-2-i][j] = tmp_od2[i][j] / (z0[i+1]-z0[i]);
  }

  free(tmp_od);
  free(tmp_wght);

  for (i=0; i<nlev-1; i++)
    free(tmp_od2[i]);
  free(tmp_od2);

  return 0;
}




/*****************************************************************/
/* Calculate cross section for a given temperature, pressure,    */
/* and concentration; David Kratz AVHRR parameterization         */
/*****************************************************************/

static int avhrr_kratz_profile (float *temper, float *press, float *z, 
				float **dens, int nlev, 
				float co2, float ch4, float n2o, 
				float f11, float f12,
				ck_profile *ck)
{

  float *z0=NULL, *p0=NULL, *t0=NULL, *u0=NULL, *ux=NULL;

  int lc=0;

  /* allocate memory for profiles */
  z0 = calloc (nlev, sizeof(float));
  p0 = calloc (nlev, sizeof(float));
  t0 = calloc (nlev, sizeof(float));
  u0 = calloc (nlev, sizeof(float));
  ux = calloc (nlev, sizeof(float));



  /* calculate input profiles */
  for (lc=0;lc<nlev;lc++) {

    z0[nlev-lc-1] = z[lc];

    p0[nlev-lc-1] = press[lc]/1013.0;   /* convert pressure from mbar to atm */
    t0[nlev-lc-1] = temper[lc];

    u0[nlev-lc-1] = dens[MOL_H2O][lc] * H2O_MOLWGHT / NAVOGADRO * 1.0E5;
    ux[nlev-lc-1] = dens[MOL_O3] [lc] * O3_MOLWGHT  / NAVOGADRO * 1.0E5;

  } 

  /* call AVHRR function */

  /* channel 1 */

  avhrr_kratz(1, 5, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4, 
	      ck[0].crs, ck[0].weight, &(ck[0].ngauss));
  avhrr_kratz(1, 4, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4, 
	      ck[1].crs, ck[1].weight, &(ck[1].ngauss));
  avhrr_kratz(1, 3, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4, 
	      ck[2].crs, ck[2].weight, &(ck[2].ngauss));
  avhrr_kratz(1, 2, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4, 
	      ck[3].crs, ck[3].weight, &(ck[3].ngauss));
  avhrr_kratz(1, 1, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4, 
	      ck[4].crs, ck[4].weight, &(ck[4].ngauss));

  /* channel 2 */
  avhrr_kratz(2, 4, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[5].crs, ck[5].weight, &(ck[5].ngauss));
  avhrr_kratz(2, 3, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[6].crs, ck[6].weight, &(ck[6].ngauss));
  avhrr_kratz(2, 2, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[7].crs, ck[7].weight, &(ck[7].ngauss));
  avhrr_kratz(2, 1, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[8].crs, ck[8].weight, &(ck[8].ngauss));

  /* channel 3 */
  avhrr_kratz(3, 5, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[9].crs, ck[9].weight, &(ck[9].ngauss));
  avhrr_kratz(3, 4, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[10].crs, ck[10].weight, &(ck[10].ngauss));
  avhrr_kratz(3, 3, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[11].crs, ck[11].weight, &(ck[11].ngauss));
  avhrr_kratz(3, 2, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[12].crs, ck[12].weight, &(ck[12].ngauss));
  avhrr_kratz(3, 1, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[13].crs, ck[13].weight, &(ck[13].ngauss));

  /* channel 4 */
  avhrr_kratz(4, 1, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[14].crs, ck[14].weight, &(ck[14].ngauss));

  /* channel 5 */
  avhrr_kratz(5, 1, z0, p0, t0, u0, ux, nlev, co2, n2o, f11, f12, ch4,
	      ck[15].crs, ck[15].weight, &(ck[15].ngauss));


  return 0; /* ok */
}




int sbdart_profile (float *temper, float *press, float *z, 
		    float *h2o, float *o3, float *dt, float *ssa, int nlev, 
		    float sza, float lambda, int iv, ck_profile *ck)
{

#if HAVE_SBDART

  int lc=0, ig=0;
  float wl = lambda / 1000.0;   /* convert from nm to micron */
  
  /* get mxly from Fortran/C include file */
  #include <sbtaugas.inc>

  void F77_FCN (sbtaugas) (int *nlev, float *zi, float *pi, float *ti,
			   float *wh, float *wo, float *sc,
			   float *wl, float *sza,
			   float *dtau, float *wght, int *npass);


  float *zi = calloc (nlev, sizeof (float));   /* inverted altitude profile    */
  float *ti = calloc (nlev, sizeof (float));   /* inverted temperature profile */
  float *pi = calloc (nlev, sizeof (float));   /* inverted pressure profile    */
	    
  float *wh = calloc (nlev, sizeof (float));   /* water vapour concentration   */
  float *wo = calloc (nlev, sizeof (float));   /* ozone concentration          */

  float *sc = calloc (nlev, sizeof (float));   /* scattering optical depth     */
  
  float dtau[MXLY*NK], wght[NK];
  int npass=0;

  float **dtauc=NULL;

  for (lc=0; lc<nlev; lc++) {
    zi[lc] = z      [nlev-lc-1];
    pi[lc] = press  [nlev-lc-1];
    ti[lc] = temper [nlev-lc-1];
    wh[lc] = h2o    [nlev-lc-1] * 1.0e6 * (M_H2O * AMU*1000.0);
    wo[lc] = o3     [nlev-lc-1] * 1.0e6 * (M_O3  * AMU*1000.0);
  }


  /* optical thickness is defined per layer, */
  /* not per level as the above quantities   */
  for (lc=0; lc<nlev-1; lc++)
    sc[lc] = dt [nlev-lc-2] * ssa [nlev-lc-2];
  

  /* ??? pass scattering optical thickness to sbtaugas() */
  /* ??? to correct in depthscl.f                        */

  F77_FCN (sbtaugas) (&nlev, zi, pi, ti, wh, wo, sc, &wl, &sza, 
		      dtau, wght, &npass);


  dtauc = fortran2c_2D_float_ary (NK, MXLY, dtau);

  /* copy data to final destination */
  ck[iv].ngauss = npass;
  
  for (ig=0; ig<npass; ig++) {
    
    /* copy data to final destination */
    ck[iv].weight[ig] = wght[ig];
    
    /* ??? ignoring optical thickness in 1st (uppermost) level */
    /* ??? what does that mean anyway ?                        */
    for (lc=0; lc<nlev-1; lc++)
      ck[iv].crs[lc][ig] = dtauc[ig][lc+1] / (z[lc]-z[lc+1]);
  }

  free (zi); free (pi); free (ti); free (wh); free (wo); free (sc);
  ASCII_free_float (dtauc, NK);

  return 0;
#else
  fprintf (stderr, "Error, SBDART gas absorption not available.\n");
  return -1;
#endif
}
  



/*****************************************************************/
/* Read cross section profile                                    */
/*****************************************************************/

static int generic_profile (char *filename, float *z, int nlev, int *nwght,
			    ck_profile *ck)
{
  int status=0;
  int iv=0, ig=0, lu=0;
  double deltaz=0;

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  int ncid=0;
  int idd_nwvn=0, idd_nlyr=0, idd_maxc=0;
  int id_nc=0, id_wght=0, id_k=0;

  int start[1] = {0};
  int count[1] = {0};

  int wghtstart[2] = {0,0};
  int wghtcount[2] = {0,0};

  int crsstart[3] = {0,0,0};
  int crscount[3] = {0,0,0};

  int nwvn=0, nlyr=0, maxc=0;

  int *nc=NULL;

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }


  /* get dimension id for "nwvn" */
  status = nc_inq_dimid (ncid, "nwvn", &idd_nwvn);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nwvn" */
  status = nc_inq_dimlen (ncid, idd_nwvn, &nwvn);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }



  /* get dimension id for "nlyr" */
  status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlyr" */
  status = nc_inq_dimlen (ncid, idd_nlyr, &nlyr);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  if (nlyr != nlev-1) {
    fprintf (stderr, "Error, different number of layers in atmospheric profile (%d)\n", nlev-1);
    fprintf (stderr, "and absorption coefficient profile %s (%d)\n", filename, nlyr);
    return -1;
  }

  /* get dimension id for "maxc" */
  status = nc_inq_dimid (ncid, "maxc", &idd_maxc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "maxc" */
  status = nc_inq_dimlen (ncid, idd_maxc, &maxc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  count[0] = nwvn;
  
  /* allocate memory for fields */
  nc    = calloc (nwvn, sizeof(int));


  /* get variable id for "nc" */
  status = nc_inq_varid (ncid, "nc", &id_nc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* read "nc" */
  status = nc_get_vara_int (ncid, id_nc, start, count, nc);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
  }


  /* get variable id for "weight" */
  status = nc_inq_varid (ncid, "weight", &id_wght);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* get variable id for "k" */
  status = nc_inq_varid (ncid, "k", &id_k);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* Loop over bands */
  for (iv=0; iv<nwvn; iv++) {
    
    /* copy data to final destination */
    ck[iv].ngauss = nc[iv];
    
    wghtstart[0] = iv;
    wghtstart[1] = 0;

    wghtcount[0] = 1;
    wghtcount[1] = nwght[iv+1];
    
    status = nc_get_vara_double (ncid, id_wght, wghtstart, wghtcount, ck[iv].weight);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    

    for (lu=0; lu<nlev-1; lu++) {

      deltaz = (z[lu] - z[lu+1]);

      crsstart[0]=iv;
      crsstart[1]=lu;
      crsstart[2]=0;
      
      crscount[0]=1;
      crscount[1]=1;
      crscount[2]=nwght[iv+1];
      
      status = nc_get_vara_double (ncid, id_k, crsstart, crscount, ck[iv].crs[lu]);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading %s\n", status, filename);
	return status;
      }
      
      for (ig=0; ig<nwght[iv+1]; ig++)
	ck[iv].crs[lu][ig] /= deltaz;
    }
  }



  free (nc);

  nc_close(ncid);

#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif
  
  return 0;  /* ok */
}








/*************************************************************************/
/* Setup correlated-k structures ck and crs_ck for given profiles.       */
/* ck holds the general information (wavelengths, number of weighting    */
/* points, ...) while crs_ck holds the detailed cross section            */
/* profiles for each component and wavelength.                           */
/*************************************************************************/

int crs_ck (ck_struct *ck, 
	    float *lambda, int n_lambda, 
	    float *temp, float *press, float *z, float **dens, int nlev,
	    float *mixing_ratio, int h2ocont, float *sza,
	    char *path, char *filename, int first, crs_ck_out_struct *crs_ck)
{
  
  int is=0, it=0, iv=0, status=0;
  double pressure=0, temperature=0, density=0, dz=0;
  
  float *dtauc_dummy = calloc (nlev, sizeof(float));
  float *ssalb_dummy = calloc (nlev, sizeof(float));



  /* Allocate memory for crs_ck */
  if (first) {
    switch(ck->scheme) {
    case CK_KATO:
      
      /* absorption by O3, O2, CO2, and H2O */
      crs_ck->profile = calloc(KATO_COMPONENTS, sizeof(ck_profile *));
      for (is=0; is<KATO_COMPONENTS; is++) 
	crs_ck->profile[is] = calloc (n_lambda, sizeof(ck_profile));

      break;

    case CK_FU:
      
      /* need only one absorption cross section; band overlap is handled internally */
      crs_ck->profile = calloc(FU_COMPONENTS, sizeof(ck_profile *));
      for (is=0; is<FU_COMPONENTS; is++) 
	crs_ck->profile[is] = calloc (n_lambda, sizeof(ck_profile));
      break;

    case CK_AVHRR_KRATZ:
      
      /* need only one absorption cross section; band overlap is handled internally */
      crs_ck->profile = calloc(AVHRR_KRATZ_COMPONENTS, sizeof(ck_profile *));
      for (is=0; is<AVHRR_KRATZ_COMPONENTS; is++) 
	crs_ck->profile[is] = calloc (n_lambda, sizeof(ck_profile));
      break;

    case CK_GENERIC:

      /* need only one absorption cross section; band overlap is handled internally */
      crs_ck->profile = calloc(GENERIC_COMPONENTS, sizeof(ck_profile *));
      for (is=0; is<GENERIC_COMPONENTS; is++) 
	crs_ck->profile[is] = calloc (n_lambda, sizeof(ck_profile));
      break;

    case CK_SBDART:

      /* need only one absorption cross section; band overlap is handled internally */
      crs_ck->profile = calloc(SBDART_COMPONENTS, sizeof(ck_profile *));
      for (is=0; is<SBDART_COMPONENTS; is++) 
	crs_ck->profile[is] = calloc (n_lambda, sizeof(ck_profile));
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", ck->scheme);
      return -1;

      break;
    }
  }
    


    
  switch(ck->scheme) {
  case CK_KATO:
    
    /* loop over bands */
    for (iv=0; iv<n_lambda; iv++) {
      
      if (first) 
	for (is=0; is<KATO_COMPONENTS; is++) {
	  crs_ck->profile[is][iv].crs     = calloc(nlev, sizeof(double *));
	  crs_ck->profile[is][iv].weight  = calloc(KATO_MAXINT, sizeof(double));
	}
      
      /* loop over layers */
      for (it=0; it<nlev-1; it++) {
	
	if (first) 
	  for (is=0; is<KATO_COMPONENTS; is++) 
	    crs_ck->profile[is][iv].crs[it] = calloc(KATO_MAXINT, sizeof(double));
	
	/* average parameters over layers */
	pressure    = 0.5 * (press[it]    + press[it+1]);     /* [mbar] */
	temperature = 0.5 * (temp[it]     + temp[it+1]);      /* [K]    */
	dz          =       (z[it]        - z[it+1])*1e5;     /* [cm]   */
	
	
	
	/* finally, calculate absorption coefficients */
	
	/* CO2 */
	density     = 0.5 * (dens[MOL_CO2][it] + dens[MOL_CO2][it+1]);
	status = kato_crs (pressure, temperature, density, dz,
			   ck, iv+1, KATO_CO2, path,
			   crs_ck->profile[KATO_CO2-1][iv].crs[it],
			   crs_ck->profile[KATO_CO2-1][iv].weight,
			   &(crs_ck->profile[KATO_CO2-1][iv].ngauss));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d calculating absorption cross section of CO2\n", status);
	  return status;
	}
	
	/* O3 */
	density     = 0.5 * (dens[MOL_O3][it] + dens[MOL_O3][it+1]);
	status = kato_crs (pressure, temperature, density, dz, 
			   ck, iv+1, KATO_O3, path,
			   crs_ck->profile[KATO_O3-1][iv].crs[it],
			   crs_ck->profile[KATO_O3-1][iv].weight,
			   &(crs_ck->profile[KATO_O3-1][iv].ngauss));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d calculating absorption cross section of O3\n", status);
	  return status;
	}
	
	
	/* O2 */
	density     = 0.5 * (dens[MOL_O2][it] + dens[MOL_O2][it+1]);
	status = kato_crs (pressure, temperature, density, dz,
			   ck, iv+1, KATO_O2, path,
			   crs_ck->profile[KATO_O2-1][iv].crs[it],
			   crs_ck->profile[KATO_O2-1][iv].weight,
			   &(crs_ck->profile[KATO_O2-1][iv].ngauss));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d calculating absorption cross section of O2\n", status);
	  return status;
	}
	
	
	/* H2O */
	density     = 0.5 * (dens[MOL_H2O][it] + dens[MOL_H2O][it+1]);
	status = kato_crs (pressure, temperature, density, dz,
			   ck, iv+1, KATO_H2O, path,
			   crs_ck->profile[KATO_H2O-1][iv].crs[it],
			   crs_ck->profile[KATO_H2O-1][iv].weight,
			   &(crs_ck->profile[KATO_H2O-1][iv].ngauss));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d calculating absorption cross section of H2O\n", status);
	  return status;
	}
      }
    }

    break;
    
  case CK_FU:
    
    if (first) {
      for (is=0; is<FU_COMPONENTS; is++) {
	for (iv=0; iv<n_lambda; iv++) {
	  crs_ck->profile[is][iv].crs    = calloc(nlev, sizeof(double *));
	  crs_ck->profile[is][iv].weight = calloc(FU_MAXINT, sizeof(double));
	    
	  for (it=0; it<nlev; it++)
	    crs_ck->profile[is][iv].crs[it] = calloc(FU_MAXINT, sizeof(double));
	}
      }
    }
    
    status = fu_profile (temp, press, z, dens, nlev, h2ocont,
			 mixing_ratio[MX_CO2],
			 mixing_ratio[MX_CH4], 
			 mixing_ratio[MX_N2O], 
			 mixing_ratio[MX_F11], 
			 mixing_ratio[MX_F12], 
			 mixing_ratio[MX_F22], 
			 crs_ck->profile[0]);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by fu_profile()\n", status);
      return status;
    }
    
    break;
      
  case CK_AVHRR_KRATZ:
    
    if (first) {
      for (is=0; is<AVHRR_KRATZ_COMPONENTS; is++) {
	for (iv=0; iv<n_lambda; iv++) {
	  crs_ck->profile[is][iv].crs    = calloc(nlev, sizeof(double *));
	  crs_ck->profile[is][iv].weight = calloc(AVHRR_KRATZ_MAXINT, sizeof(double));
	  
	  for (it=0; it<nlev; it++)
	    crs_ck->profile[is][iv].crs[it] = calloc(AVHRR_KRATZ_MAXINT, sizeof(double));
	}
      }
    }
    
    
    status = avhrr_kratz_profile (temp, press, z, dens, nlev,
				  mixing_ratio[MX_CO2],
				  mixing_ratio[MX_CH4], 
				  mixing_ratio[MX_N2O], 
				  mixing_ratio[MX_F11],
				  mixing_ratio[MX_F12],
				  crs_ck->profile[0]);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by avhrr_kratz_profile()\n", status);
      return status;
    }
    
    break;
      

  case CK_SBDART:


    /* loop over bands */
    for (iv=0; iv<n_lambda; iv++) {
      
      if (first) 
	for (is=0; is<SBDART_COMPONENTS; is++) {
	  crs_ck->profile[is][iv].crs     = calloc(nlev, sizeof(double *));
	  crs_ck->profile[is][iv].weight  = calloc(SBDART_MAXINT, sizeof(double));

	  for (it=0; it<nlev-1; it++)
	    crs_ck->profile[is][iv].crs[it] = calloc(SBDART_MAXINT, sizeof(double));
	}
      
      /* calculate absorption coefficients */
      
      status = sbdart_profile (temp, press, z, dens[MOL_H2O], dens[MOL_O3], 
			       dtauc_dummy, ssalb_dummy, nlev, sza[iv], lambda[iv],
                               iv, crs_ck->profile[0]);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by sbdart_profile()\n", status);
	return status;
      }
    }

    break;

  case CK_GENERIC:

    if (first) {
      for (is=0; is<GENERIC_COMPONENTS; is++) {
	for (iv=0; iv<n_lambda; iv++) {
	  crs_ck->profile[is][iv].crs    = calloc(nlev, sizeof(double *));
	  crs_ck->profile[is][iv].weight = calloc(ck->wght[GENERIC_WGHT][iv+1], sizeof(double));
	  
	  for (it=0; it<nlev; it++)
	    crs_ck->profile[is][iv].crs[it] = calloc(ck->wght[GENERIC_WGHT][iv+1], sizeof(double));
	}
      }
    }
    
    status = generic_profile (filename, z, nlev, ck->wght[GENERIC_WGHT], crs_ck->profile[0]);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by generic_profile()\n", status);
      return status;
    }
    
    break;

  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", ck->scheme);
    return -1;
    
      break;
  }

  return 0;
}




/* Rayleigh cross section for the 6 solar bands of the Fu and Liou [1992/93] */
/* parameterization; see Fu and Liou [1993], pg, 2017.                       */

double fu_rayleigh (double pressure, double temperature, int iv)
{
  /* These coefficients don't seem right, re-calculation gives  */
  /* 7.7333e-06 for the first                                   */

  double R[6] = {9.022E-6, 5.282E-7, 5.722E-8, 1.433E-8, 4.526E-9, 1.529E-9};

  
  if (iv>=6)
    return 0;
  else
    return R[iv] * pressure / temperature;
}



/* routines to read the Kato et al. tables; translated from */
/* Seji Kato's fortran program using f2c                    */

static int x_section (int ispcs, int ipress, int itemp, 
		      double presslog, double temp, int iband, char *path, 
		      int *ngauss, double *xabs, double *weight)
{
#if HAVE_LIBNETCDF
  static float xgrid[17] = { 3.02119, 2.92117, 2.82086, 2.72099, 
			     2.62118, 2.52114, 2.42160, 2.22531, 
			     2.02531, 1.82543, 1.62531, 1.42488, 
			     1.22531, 1.02531, 0.00000, -1.0000,
			     -2.0000 };
  
  static float ygrid[11] = { 120.0, 140.0, 160.0, 180.0, 
			     200.0, 220.0, 240.0, 260.0, 
			     280.0, 300.0, 320.0 };

  
  static int c__1 = 1;
  static int c__2 = 2;
  static int c__0 = 0;

  /* netcdf Fortran functions */
  extern int ncvid_();
  extern int ncopn_();
  extern int ncvgt_();
  extern int ncclos_(), ncpopt_();

  static int w_id__, err_code__, irho, cdfid, i, k;
  static char sband[5];
  static char xfile[FILENAME_MAX];
  static double frcnx, frcny;
  static double xsect[80]	/* was [4][20] */;
  static int count[3], start[3];
  static double xsectt[20];
  static int xabs_id__;
  


/* This subroutine is written by Seiji Kato */
/*          The Pannsylvania State University */
/*          Meteorology Department */
/*          e-mail kato@essc.psu.edu */
/*          25 March 1996 */
/* This subroutine returns absorption cross section in cm^2 per molecule */
/* at n-gaussian points at given pressure, temperature. */
/* The unit of the cross section is cm^2 per molecule. */
/* Input */
/* ipress */
/* itemp */
/*presslog    log10 of pressure (mb), which you want to obtain cross sections.*/
/*temp        log10 of temperature (K), which you want to obtain cross sections.*/
/* iband       Bnad number based on the attached table. */

/*ngauss      Number of the Gaussian quadratures based on the attached table*/
/* Output */
/* xabs        Absorption cross-sections at the Gaussian quadrature points
 */
/*             in cm^2 per molecule. */
/* weight      Gaussian weights */
/* This subroutine returns absorption cross section in cm^2 per molecule 
*/
/*at n-gaussian points at given pressure, temperature, and number concentr
ation*/
/* of species. The species has to be either 'CO2', 'O3', or 'O2'. */
/* ispcs = 1  :CO2 */
/* ispcs = 2  :O3 */
/* ispcs = 3  :O2 */
/* ispcs = 4  ;H2O */
/* The unit of the cross section is cm^2 per molecule. */


  /* Parameter adjustments */
  --weight;
  --xabs;
  
  /* Function Body */
  /* temperature grid */
  

  strcpy (xfile, path);
  strcat (xfile, "correlated_k/kato/cross_section.table.");
  
  sprintf (sband, "%d", iband);
  
  switch (ispcs) {
  case KATO_CO2:
    strcat(xfile, "CO2.");
    break;
  case KATO_O3:
    strcat (xfile, "_O3.");
    break;
  case KATO_O2:
    strcat (xfile, "_O2.");
    break;
  case KATO_H2O:
    strcat (xfile, "H2O.");
    break;
  default:
    fprintf (stderr, "     Wrong species\n");
  }
  
  if (iband <= 9)
    strcat(xfile, "0");
  
  strcat(xfile, sband);
  strcat(xfile, ".cdf");
  

  /* I want error message, but do not wish error to */
  /* be fatal. */
  ncpopt_(&c__2);
  
  /* open the netcdf file */
  cdfid = ncopn_(xfile, &c__0, &err_code__, strlen(xfile));

  /* get the id for gaussian weights */
  w_id__ = ncvid_(&cdfid, "weight", &err_code__, 6L);
    
  /* c get the ids of cross sections */
  xabs_id__ = ncvid_(&cdfid, "cross_section", &err_code__, 13L);
  
  /* get weights */
  ncvgt_(&cdfid, &w_id__, &c__1, ngauss, &weight[1], &err_code__);

  /* get cross sections */
  start[0] = ipress;
  start[1] = itemp;
  start[2] = 1;
  
  count[0] = 1;
  count[1] = 1;
  count[2] = *ngauss;

  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
  
  for (i=1; i<=*ngauss; ++i) 
    xsect[(i << 2) - 4] = xsectt[i - 1];
  
  start[0] = ipress + 1;
  start[1] = itemp;
  start[2] = 1;
  
  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
  
  for (i=1; i<=*ngauss; ++i)
    xsect[(i << 2) - 3] = xsectt[i - 1];
  
  start[0] = ipress;
  start[1] = itemp + 1;
  start[2] = 1;
  
  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
  
  for (i=1; i<=*ngauss; ++i)
    xsect[(i << 2) - 1] = xsectt[i - 1];

  
  start[0] = ipress + 1;
  start[1] = itemp + 1;
  start[2] = 1;
  
  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
  
  for (i=1; i<=*ngauss; ++i)
    xsect[(i << 2) - 2] = xsectt[i - 1];
  
  
  ncclos_(&cdfid, &err_code__);
  
  /* Checking interpolation region */
  for (i=1; i<=8; ++i)
    if (xsect[i-1] <= 0.0) {
      fprintf (stderr, "Warning, the condition is out of the envelope  %d %d %d\n",
	       ipress, itemp, irho);
      
      fprintf (stderr, "temp = %f, presslog = %f\n", temp, presslog);
    }

  /* Do two-dimensional interpolation on pressure and temperature plane. */
  frcnx = (presslog - xgrid[ipress - 1]) / (xgrid[ipress] - xgrid[ipress - 1]);
  frcny = (temp - ygrid[itemp - 1]) / (ygrid[itemp] - ygrid[itemp - 1]);
  
  for (k = 1; k <= *ngauss; ++k) {
    xabs[k] = (1.0 - frcnx) * (1.0 - frcny) * xsect[(k << 2) - 4] + frcnx * 
      (1.0 - frcny) * xsect[(k << 2) - 3] + frcnx * frcny * 
      xsect[(k<< 2) - 2] + (1.0 - frcnx) * frcny * xsect[(k << 2) - 1];
  }
#endif  
  return -1;
}




static int x_section_h2o (int *ipress, int *itemp, int *irho,
			  double *presslog, double *temp, 
			  double *rholog, int *iband, char *path, 
			  int *ngauss, double *xabs, double *weight)
{
  /* Initialized data */
#if HAVE_LIBNETCDF
  
  static float xgrid[17] = { 3.02119, 2.92117, 2.82086, 2.72099, 
			     2.62118, 2.52114, 2.42160, 2.22531, 
			     2.02531, 1.82543, 1.62531, 1.42488, 
			     1.22531, 1.02531, 0.00000, -1.0000,
			     -2.0000 };
  
  static float ygrid[11] = { 120.0, 140.0, 160.0, 180.0, 
			     200.0, 220.0, 240.0, 260.0, 
			     280.0, 300.0, 320.0 };
  
  static float zgrid[11] = { 24.1761, 23.6721, 23.1761, 22.6721,
                             22.1761, 21.6721, 21.1761, 20.6721,
                             20.1761, 19.6721, 19.1761 };
  

  /* Local variables */
  static int c__1 = 1;
  static int c__2 = 2;
  static int c__0 = 0;

  /* netcdf Fortran functions */
  extern int ncvid_();
  extern int ncopn_();
  extern int ncvgt_();
  extern int ncclos_(), ncpopt_();

  static int w_id__, err_code__, cdfid, i, k;
  static char sband[5];
  static char xfile[FILENAME_MAX];
  static double frcnx, frcny;
  static double xsect[160]	/* was [8][20] */;
  static int count[4], start[4];
  static double ansyrl, xsectt[20], ansyru;
  static int xabs_id__;


  /* This subroutine is written by Seiji Kato */
  /*          The Pannsylvania State University */
  /*          Meteorology Department */
  /*          e-mail kato@essc.psu.edu */
  /*          25 March 1996 */

    /* Parameter adjustments */
  --weight;
  --xabs;

  /* Function Body */

  strcpy (xfile, path);
  strcat (xfile, "correlated_k/kato/cross_section.table.H2O.");
  
  if (*iband <= 9)
    strcat(xfile, "0");
  
  sprintf (sband, "%d", *iband);

  strcat(xfile, sband);
  strcat(xfile, ".cdf");


  ncpopt_(&c__2);

  cdfid = ncopn_(xfile, &c__0, &err_code__, strlen(xfile));
  w_id__ = ncvid_(&cdfid, "weight", &err_code__, 6L);

  xabs_id__ = ncvid_(&cdfid, "cross_section", &err_code__, 13L);
  ncvgt_(&cdfid, &w_id__, &c__1, ngauss, &weight[1], &err_code__);

  start[0] = *ipress;
  start[1] = *itemp;
  start[2] = *irho;
  start[3] = 1;

  count[0] = 1;
  count[1] = 1;
  count[2] = 1;
  count[3] = *ngauss;

  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);

  for (i=1; i<=*ngauss; ++i) {
  L1001:
    if (start[2]<=10 && i==1) {
      if (xsectt[i-1] <= 0) {
	++start[2];
	ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	goto L1001;
      } else {
	goto L1002;
      }
    } else {
      goto L1002;
    }
  L1002:
    xsect[(i << 3) - 8] = xsectt[i-1];
  }

  start[0] = *ipress + 1;
  start[1] = *itemp;
  start[2] = *irho;
  start[3] = 1;

  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);

  for (i=1; i<=*ngauss; ++i) {
  L1011:
    if (start[2]<=10 && i==1) {
      if (xsectt[i-1] <= 0) {
	++start[2];
	ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	goto L1011;
      } else {
	goto L1012;
      }
    } else {
      goto L1012;
    }

  L1012:
    xsect[(i << 3) - 7] = xsectt[i-1];
  }

  start[0] = *ipress;
  start[1] = *itemp + 1;
  start[2] = *irho;
  start[3] = 1;

  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
    
  for (i=1; i<=*ngauss; ++i) {
  L1021:
    if (start[2]<=10 && i==1) {
      if (xsectt[i-1] <= 0) {
	++start[2];
	ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	goto L1021;
      } else {
	goto L1022;
      }
    } else {
      goto L1022;
    }

  L1022:
    xsect[(i << 3) - 5] = xsectt[i-1];
  }
  start[0] = *ipress + 1;

  start[1] = *itemp + 1;
  start[2] = *irho;
  start[3] = 1;

  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);

    
  for (i=1; i<=*ngauss; ++i) {

  L1031:
    if (start[2]<=10 && i==1) {
      if (xsectt[i-1] <= 0) {
	++start[2];
	ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	goto L1031;
      } else {
	goto L1032;
      }
    } else {
      goto L1032;
    }

  L1032:
    xsect[(i << 3) - 6] = xsectt[i-1];
  }
  if (*irho <= 10) {
    start[0] = *ipress;
    start[1] = *itemp;
    start[2] = *irho + 1;
    start[3] = 1;

    ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);

	
    for (i=1; i<=*ngauss; ++i) {

    L1041:
      if (start[2]<=10 && i==1) {
	if (xsectt[i-1] <= 0) {
	  ++start[2];
	  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &
		 err_code__);
	  goto L1041;
	} else {
	  goto L1042;
	}
      } else {
	goto L1042;
      }

    L1042:
      xsect[(i << 3) - 4] = xsectt[i-1];

    }
    start[0] = *ipress + 1;
    start[1] = *itemp;
    start[2] = *irho + 1;
    start[3] = 1;

    ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);

	
    for (i=1; i<=*ngauss; ++i) {
    L1051:
      if (start[2] <= 10 && i==1) {
	if (xsectt[i-1] <= 0) {
	  ++start[2];

	  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &
		 err_code__);
	  goto L1051;
	} else {
	  goto L1052;
	}
      } else {
	goto L1052;
      }

    L1052:
      xsect[(i << 3) - 3] = xsectt[i-1];
    }
    start[0] = *ipress;
    start[1] = *itemp + 1;
    start[2] = *irho + 1;
    start[3] = 1;

    ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	
    for (i=1; i<=*ngauss; ++i) {

    L1061:
      if (start[2]<=10 && i==1) {
	if (xsectt[i-1] <= 0) {
	  ++start[2];

	  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &
		 err_code__);
	  goto L1061;
	} else {
	  goto L1062;
	}
      } else {
	goto L1062;
      }

    L1062:
      xsect[(i << 3) - 1] = xsectt[i-1];
    }
    start[0] = *ipress + 1;
    start[1] = *itemp + 1;
    start[2] = *irho + 1;
    start[3] = 1;

    ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &err_code__);
	
    for (i=1; i<=*ngauss; ++i) {

    L1071:
      if (start[2]<=10 && i==1) {
	if (xsectt[i-1] <= 0) {
	  ++start[2];
	  ncvgt_(&cdfid, &xabs_id__, start, count, xsectt, &
		 err_code__);

	  goto L1071;
	} else {
	  goto L1072;
	}
      } else {
	goto L1072;
      }

    L1072:
      xsect[(i << 3) - 2] = xsectt[i-1];
    }
    /* Checking interpolation region */
    for (i=1; i<=8; ++i) {
      if (xsect[i-1] <= 0) {
	fprintf (stderr, "Warning, the condition is out of the envelope: %d  %d  %d  %d\n",
		 *ipress, *itemp, *irho, i); 

      }
    }
    frcnx = (*presslog - xgrid[*ipress - 1]) / (xgrid[*ipress] - xgrid[*ipress - 1]);

    frcny = (*temp - ygrid[*itemp - 1]) / (ygrid[*itemp] - ygrid[*itemp - 1]);
	
    for (k = 1; k <= *ngauss; ++k) {
      ansyrl = (1.0 - frcnx) * (1.0 - frcny) * xsect[(k << 3) - 8] + 
	frcnx * (1.0 - frcny) * xsect[(k << 3) - 7] + frcnx * 
	frcny * xsect[(k << 3) - 6] + (1.0 - frcnx) * frcny * 
	xsect[(k << 3) - 5];

      ansyru = (1.0 - frcnx) * (1.0 - frcny) * xsect[(k << 3) - 4] + 
	frcnx * (1.0 - frcny) * xsect[(k << 3) - 3] + frcnx * 
	frcny * xsect[(k << 3) - 2] + (1.0 - frcnx) * frcny * 
	xsect[(k << 3) - 1];

      xabs[k] = (ansyru - ansyrl) * (*rholog - zgrid[*irho - 1]) / (
								    zgrid[*irho] - zgrid[*irho - 1]) + ansyrl;
    }
  } else {
    /* Checking interpolation region */
    for (i=1; i<=4; ++i) {
      if (xsect[i-1] <= 0) {
	fprintf (stderr, "Warning, the condition is out of the envelope: %d  %d  %d  %d\n",
		 *ipress, *itemp, *irho, i); 
      }
    }

    frcnx = (*presslog - xgrid[*ipress - 1]) / (xgrid[*ipress] - xgrid[*ipress - 1]);
      
    frcny = (*temp - ygrid[*itemp - 1]) / (ygrid[*itemp] - ygrid[*itemp - 1]);
	
    for (k = 1; k <= *ngauss; ++k) {

      xabs[k] = (1.0 - frcnx) * (1.0 - frcny) * xsect[(k << 3) - 8] + 
	frcnx * (1.0 - frcny) * xsect[(k << 3) - 7] + frcnx * 
	frcny * xsect[(k << 3) - 6] + (1.0 - frcnx) * frcny * 
	xsect[(k << 3) - 5];
    }
  }

  ncclos_(&cdfid, &err_code__);
#endif
  return 0;
}





static int hunt(float *xx, int *n, double *x, int *jlo)
{
  static int ascnd=0;
  static int jm=0, inc=0, jhi=0;
  
  /* Parameter adjustments */
  --xx;
  
  /* Function Body */
  ascnd = xx[*n] > xx[1];
  
  if (*jlo<=0 || *jlo>*n) {
    *jlo = 0;
    jhi = *n + 1;
    goto L3001;
  }
  
  inc = 1;


  if (*x>=xx[*jlo] == ascnd) {   /* Hunt up */
  L1001:
    jhi = *jlo + inc;
    
    if (jhi > *n) {
      jhi = *n + 1;
    } else if (*x >= xx[jhi] == ascnd) {
      *jlo = jhi;
      inc += inc;
      goto L1001;
    }
  } else {  /* Hunt doun */
    jhi = *jlo;
  L2001:
    *jlo = jhi - inc;
    if (*jlo < 1) {
      *jlo = 0;
    } else if (*x < xx[*jlo] == ascnd) {
      jhi = *jlo;
      inc += inc;
      goto L2001;
    }
  }

  /* Hunt is done, so begin the final bisection phase */
L3001:
  if (jhi - *jlo == 1)
    return 0;


  jm = (jhi + *jlo) / 2;

  if (*x > xx[jm] == ascnd)
    *jlo = jm;
  else
    jhi = jm;

  goto L3001;
}

