/*--------------------------------------------------------------------
 * $Id: cloud.c,v 1.14 2001/10/10 22:40:47 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/miecalc.h"
#include "../src_f/f77-uscore.h"

#if HAVE_YANG56
#include "../src_c/yang56.h"
#endif

#define EPSILON 1E-6



/************************************/
/* prototypes of internal functions */
/************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			cld_out_struct  *cld_out);

static int compare_levels (float *z1, int n1,
			   float *z2, int n2);

static int interpolate_legendre (float **legen, int *nleg, int nreff,
				 double r0, double dr,
				 double *extinc, double *albedo,
				 double reff,
				 float **coeffc, int *ncoeffc, 
				 float *ext, float *alb);

static int calloc_cld_microphysics (cld_out_struct *out, int Nlam, int Nlev);

static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer);




/**************************************************************/
/* Setup water clouds.                                        */
/**************************************************************/

int setup_wcloud (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;


  /* if water cloud microphysical properties profile specified */
  if (strlen(input.filename[FN_WCLOUD]) > 0) {

    /* Read microphysical properties from input.filename[FN_WCLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose) {
      fprintf (stderr, " ... reading water cloud properties from %s\n", 
	       input.filename[FN_WCLOUD]);
    }

    /* Read water cloud microphysical properties and apply */
    /* default conversion to optical properties            */
    status = read_and_convert_wc_file (input, output, 
				       input.filename[FN_WCLOUD], 
				       &(output->wc));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_wc_file (%s)\n", 
	       status, input.filename[FN_WCLOUD]);
      return status;
    }


    /* Read optical properties file if available */
    if (strlen(input.filename[FN_WCLOUD_PROP]) > 0) {

      if (input.wc.properties != CLDPROP_FILE && 
	  input.wc.properties != CLDPROP_MIE) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	return -1;
      }

      /* read single scattering properties from file, */
      /* allocate memory of moments                   */
      status = read_wcloud_prop (input, output);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading water cloud property file %s\n", 
		 status, input.filename[FN_WCLOUD_PROP]);
	return status;
      }
    }
  }
  else {   
    
    /* Read file holding filenames of layer cloud optical properties */    
    if (strlen(input.filename[FN_WCLOUD_FILES]) > 0) {
      
      if (input.wc.properties != CLDPROP_EXPLICIT) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	return -1;
      }

      if (input.atm.altitude != 0) {
	fprintf (stderr, "Sorry, option 'altitude' currently not implemented in combination\n");
	fprintf (stderr, "with 'wc_files'\n");
	return -1;
      }
      
      /* allocate memory for optical properties profiles */
      status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, output->atm.nlev);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }

      /* read optical properties profile, */
      /* allocate memory of moments       */

      status = read_optprop_files (input.filename[FN_WCLOUD_FILES], output->atm.nlyr,
				   output->wl.lambda_t, output->wl.nlambda_t, 
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->atm.zd,
				   output->wc.optprop.dtau, 
				   output->wc.optprop.ssa, 
				   output->wc.optprop.gg,
				   output->wc.optprop.moment, output->wc.optprop.nmom);
      if (status != 0) {
	fprintf (stderr, "Error %d reading cloud optical property files: %s\n", 
		 status, input.filename[FN_WCLOUD_FILES]);
	return status;
      }
      
      
      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->atm.nlyr; lc++)
	  /* BBMM  if (output->wc.nmom[iv][lc] > output->atm.nmom ) */
	  /* BBMM    output->atm.nmom = output->wc.nmom[iv][lc];     */
	  if (output->wc.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	    output->atm.nmom = output->wc.optprop.nmom[iv][lc]-1;
    }
    else {    /* if neither of the cloud property files defined */

      /* allocate memory, even if no cloud is defined; */
      /* it may be required by other functions.        */
      
      status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, output->atm.nlev);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }
    }
  }


  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.wc, output->atm, output->wl, &(output->wc));
  if (status!=0) {
    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
	     status);
    return status;
  }
    

  return 0;
}




/**************************************************************/
/* Setup ice clouds.                                          */
/**************************************************************/

int setup_ice (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;


  /* if ice cloud microphysical properties profile specified */
  if (strlen(input.filename[FN_ICLOUD]) > 0) {

    /* Read microphysical properties from input.filename[FN_ICLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose)
      fprintf (stderr, " ... reading ice cloud microphysical properties from %s\n", 
	       input.filename[FN_ICLOUD]);
    
    /* Read ice cloud microphysical properties and apply */
    /* default conversion to optical properties          */
    status = read_and_convert_ic_file (input, output, 
				       input.filename[FN_ICLOUD], 
				       &(output->ic));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_ic_file (%s)\n", 
	       status, input.filename[FN_ICLOUD]);
      return status;
    }

    /* Read optical properties file if available */
    if (strlen(input.filename[FN_ICLOUD_PROP]) > 0) {

      if (input.ic.properties != CLDPROP_FILE && 
	  input.ic.properties != CLDPROP_MIE) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	return -1;
      }

      status = read_icloud_prop (input, output);

      if (status!=0) {
	fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		 status, input.filename[FN_ICLOUD_PROP]);
	return status;
      }
    }
  }
  else {

    /* Read file holding filenames of layer cloud optical properties */
    if (strlen(input.filename[FN_ICLOUD_FILES]) > 0) {
      
      if (input.ic.properties != CLDPROP_EXPLICIT) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	return -1;
      }

      if (input.atm.altitude != 0) {
	fprintf (stderr, "Sorry, option 'altitude' currently not implemented in combination\n");
	fprintf (stderr, "with 'ic_files'\n");
	return -1;
      }
      
      /* allocate memory for optical properties profiles */
      status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, output->atm.nlev);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }
      
      /* read optical properties profile, */
      /* allocate memory of moments       */

      status = read_optprop_files (input.filename[FN_ICLOUD_FILES], output->atm.nlyr,
				   output->wl.lambda_t, output->wl.nlambda_t,
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->atm.zd,
				   output->ic.optprop.dtau, 
				   output->ic.optprop.ssa, 
				   output->ic.optprop.gg,
				   output->ic.optprop.moment, output->ic.optprop.nmom);
      if (status != 0) {
	fprintf (stderr, "Error %d reading cloud optical property files: %s\n", 
		 status, input.filename[FN_ICLOUD_FILES]);
	return status;
      }
    
      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->atm.nlyr; lc++)
	  /* BBMM if (output->ic.nmom[iv][lc] > output->atm.nmom ) */
	  /* BBMM   output->atm.nmom = output->ic.nmom[iv][lc];     */
	  if (output->ic.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	    output->atm.nmom = output->ic.optprop.nmom[iv][lc]-1;
    }
    else { /* if neither of the cloud property files defined */
      
      /* allocate memory, even if no cloud is defined; */
      /* it may be required by other functions.        */
      
      status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, output->atm.nlev);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }
    }
  }


  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.ic, output->atm, output->wl, &(output->ic));
  if (status!=0) {
    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
	     status);
    return status;
  }
  
  return 0;
}




/**************************************************************/
/* Shift clouds the same way as the atmosphere                */
/* (see shift_atmosphere).                                    */
/**************************************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			cld_out_struct  *cld_out)
{
  int lev=0;

  /* calculate shifted levels (altitude now in km) */
  for (lev=0; lev<cld_out->nlev; lev++)
    cld_out->zd[lev] = cld_out->zd[lev] - atm_inp->altitude;

  /* check for strange things */
  for (lev=0; lev<cld_out->nlev; lev++)
    if (cld_out->zd[lev]<0 && cld_out->microphys.lwc[lev]>0) {
      fprintf (stderr, "Error: the first cloud level is below the surface; you need\n");
      fprintf (stderr, "to change the cloud profile!\n");
      return -1;
    }


  /* determine the first level below 0 */
  lev = cld_out->nlev;
  while (cld_out->zd[--lev]<=0) {
    if (lev<=1) {
      fprintf (stderr, "Sorry, ALTITUDE is higher than the uppermost layer in the\n");
      fprintf (stderr, "water cloud profile.\n");
      return -1;
    }
  }

  /* number of levels is probably smaller now */
  cld_out->nlev = lev+2;   /* ??? */

  /* set the last level to 0 */
  cld_out->zd[cld_out->nlev-1] = 0.0; 


  return 0;
}




/*********************************/
/* compare two altitude grids    */
/*********************************/

static int compare_levels (float *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    fprintf (stderr, "%d vs %d\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<n1; kc++)
    if (fabs((float) (z1[kc]) - z2[kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}




static int calloc_cld_microphys_struct (cld_microphys_struct *out, int Nlev) 
{
  out->lwc  = calloc(Nlev, sizeof(float));
  out->effr = calloc(Nlev, sizeof(float));
  out->alloc = 1;

  return 0;
}


static int calloc_cld_optprop_struct (cld_optprop_struct *out, int Nlam, int Nlev) 
{
  int iv=0, lc=0;

  out->nlam = Nlam;
  
  out->dtau = calloc(Nlam, sizeof (float *));
  out->gg   = calloc(Nlam, sizeof (float *));
  out->ssa  = calloc(Nlam, sizeof (float *));

  /* parameters for double HG */
  out->g1   = calloc(Nlam, sizeof (float *));
  out->g2   = calloc(Nlam, sizeof (float *));
  out->f    = calloc(Nlam, sizeof (float *)); /*agf*/

  
  for (iv=0; iv<Nlam; iv++) {
    out->dtau[iv] = calloc (Nlev, sizeof(float));
    out->gg  [iv] = calloc (Nlev, sizeof(float));
    out->ssa [iv] = calloc (Nlev, sizeof(float));
    out->g1  [iv] = calloc (Nlev, sizeof(float)); /*agf*/
    out->g2  [iv] = calloc (Nlev, sizeof(float));
    out->f   [iv] = calloc (Nlev, sizeof(float));
  }
  
  out->nmom   = calloc (Nlam, sizeof (int *));
  out->moment = calloc (Nlam, sizeof (float **)); 
  for (iv=0; iv<Nlam; iv++) {
    out->nmom[iv]   = calloc (Nlev, sizeof (int ));
    out->moment[iv] = calloc (Nlev, sizeof (float *));
    
    /* initialize nmom to -1; nmom==0 means an isotropic phase function */
    for (lc=0; lc<Nlev; lc++)
      /* BBMM out->nmom[iv][lc] = -1; */
      out->nmom[iv][lc] = 0;
  }
  
  out->alloc = 1;

  return 0;
}


static int calloc_cld_ssprop_struct (cld_ssprop_struct *out, int Nlam) 
{

  out->extinc = calloc (Nlam, sizeof(double *));
  out->albedo = calloc (Nlam, sizeof(double *));
  out->legen  = calloc (Nlam, sizeof(float **));
  out->nleg   = calloc (Nlam, sizeof(int *));
  out->r0     = calloc (Nlam, sizeof(double));
  out->dr     = calloc (Nlam, sizeof(double));
  out->nreff  = calloc (Nlam, sizeof(int));
  
  out->alloc  = 1;

  return 0;
}



/***************************************/
/* allocate memory for a cld_out_struct */
/***************************************/

int calloc_cld_out (cld_out_struct *out, int Nlam, int Nlev) 
{
  int status=0;

  out->zd     = calloc(Nlev, sizeof(float));
  out->nlev   = Nlev;
  out->newsiz = 1;


  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }
  
  status = calloc_cld_optprop_struct (&(out->optprop), Nlam, out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud optical properties\n", status);
    return status;
  }

  status = calloc_cld_ssprop_struct (&(out->ssprop), Nlam);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud single scattering properties\n", 
	     status);
    return status;
  }

  return 0;
}



/***********************************************************/
/* allocate memory for a cld_out_struct, microphysics only */
/***********************************************************/

static int calloc_cld_microphysics (cld_out_struct *out, int Nlam, int Nlev) 
{
  int status=0;

  out->zd     = calloc(Nlev, sizeof(float));
  out->nlev   = Nlev;
  out->newsiz = 1;
  
  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }

  return 0;
}





/************************************************/
/* copy contents of a cld_out_struct to another; */
/* NO memory allocation is done here!!!         */
/************************************************/

int cp_cld_out (cld_out_struct *target, cld_out_struct source, int copy_ssprop)
{
  int iv=0, ir=0, k=0, lev=0;

  if (target->nlev != source.nlev) {
    fprintf (stderr, "Error: cp_cld_out(), nlev differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.nlev, target->nlev);
    return -1;
  }

  target->newsiz = source.newsiz;

  /* copy microphysical properties */
  for (lev=0; lev<source.nlev; lev++) {
    target->zd[lev] = source.zd[lev];

    if (target->microphys.alloc) {
      target->microphys.lwc  [lev] = source.microphys.lwc  [lev];
      target->microphys.effr [lev] = source.microphys.effr [lev];
    }
  }

  /* copy single scattering properties */
  if (source.ssprop.alloc && copy_ssprop)  {
    target->ssprop.alloc = source.ssprop.alloc;

    for (iv=0; iv<source.optprop.nlam; iv++) {
      target->ssprop.nreff[iv] = source.ssprop.nreff[iv];
      target->ssprop.r0[iv]    = source.ssprop.r0[iv];
      target->ssprop.dr[iv]    = source.ssprop.dr[iv];

      for (ir=0; ir<source.ssprop.nreff[iv]; ir++) {
	target->ssprop.extinc [iv][ir] = source.ssprop.extinc [iv][ir];
	target->ssprop.albedo [iv][ir] = source.ssprop.albedo [iv][ir];

	target->ssprop.nleg [iv][ir]   = source.ssprop.nleg [iv][ir];
	
	for (k=0; k<source.ssprop.nleg[iv][ir]; k++)
	  target->ssprop.legen [iv][ir][k] = source.ssprop.legen [iv][ir][k];
      }
    }
  }
  

  /* copy optical properties */
  if (target->optprop.alloc && source.optprop.alloc)

    /* check if number of wavelength bands equal */
    if (target->optprop.nlam != source.optprop.nlam) {
      fprintf (stderr, "Error: cp_cld_out(), nlam differs between source and target\n");
      fprintf (stderr, "       source: %d,  target: %d\n", 
	       source.optprop.nlam, target->optprop.nlam);
      return -1;
    }
  
    for (lev=0; lev<source.nlev-1; lev++) {
      for (iv=0; iv<source.optprop.nlam; iv++) {
	target->optprop.dtau [iv][lev] = source.optprop.dtau [iv][lev];
	target->optprop.gg   [iv][lev] = source.optprop.gg   [iv][lev];
	target->optprop.ssa  [iv][lev] = source.optprop.ssa  [iv][lev];
      }
    }
  
  return 0;
}




/* Interpolate the Legendre coefficients which are given on a grid of  */
/* effective radii r0, r0 + dr, ..., r0 + (nreff-1) * dr to a given    */
/* radius reff; memory for the array of Legendre coefficients coeffc[] */
/* is allocated automatically.                                         */

static int interpolate_legendre (float **legen, int *nleg, int nreff,
				 double r0, double dr,
				 double *extinc, double *albedo,
				 double reff,
				 float **coeffc, int *ncoeffc, float *ext, float *alb)
{
  int i=0, k=0;
  int lower=0, upper=0;
  int min=0, max=0;

  double reffl=0;

  if (reff<r0 || reff> r0 + (double) (nreff-1) * dr) {
    fprintf (stderr, "Error, reff = %f out of range (%f - %f)\n",
	     reff, r0, r0 + (double) (nreff-1) * dr);
    return -1;
  }

  lower = (int) ((reff - r0) / dr);
  upper = lower+1;
  
  reffl = r0 + (double) lower * dr;

  if (lower==nreff-1) {   /* upper boundary value */
    *coeffc  = calloc (nleg[nreff-1], sizeof(float));
    *ncoeffc = nleg[nreff-1];

    for (k=0; k<*ncoeffc; k++)
      (*coeffc)[k] = legen[nreff-1][k];
    
    *alb = albedo[lower];
    *ext = extinc[lower];

    return 0;
  }

  /* otherwise interpolation required */
  min = (nleg[lower] < nleg[upper] ? nleg[lower] : nleg[upper]);
  max = (nleg[lower] > nleg[upper] ? nleg[lower] : nleg[upper]);
  
  *ncoeffc = max;
  *coeffc  = calloc (*ncoeffc, sizeof(float));

  for (i=0; i<min; i++)
    (*coeffc)[i] = legen[lower][i] + (reff - reffl) / dr * (legen[upper][i] - legen[lower][i]);
    
  for (i=min; i<*ncoeffc; i++)
    if (nleg[lower] < nleg[upper])
      (*coeffc)[i] = (reff - reffl) / dr * (legen[upper][i]);
    else
      (*coeffc)[i] = legen[lower][i] - (reff - reffl) / dr * legen[lower][i];

  
  /* extinction and single scattering albedo */
  *ext = extinc[lower] + (reff - reffl) / dr * (extinc[upper] - extinc[lower]);
  *alb = albedo[lower] + (reff - reffl) / dr * (albedo[upper] - albedo[lower]);

  return 0;
}



/* Read cloud data from file; microphysical properties are expected  */
/* in three columns:                                                 */
/*   column WCZD:    z [km]                                          */
/*   column WCLWC:   lwc [g / m3]                                    */
/*   column WCEFFR:  effective radius [micron]                       */
/* Memory for microphysical and optical properties is allocated      */
/* automatically. This function can be used for ice and water clouds */
/* because all output is written to cld_out_struct *cld which can    */
/* either be output->wc or output->ic                                */

int read_cld_file (input_struct input, output_struct *output, 
		   char *filename, cld_out_struct *cld,
		   int optical) 
{
  int lev=0, rows=0, min_columns=0, max_columns=0, status=0;
  float **cld_data=NULL;
  
  
  status = ASCII_file2float (filename,
			     &rows, &max_columns, &min_columns, 
			     &cld_data);
  if (status!=0) {
    fprintf (stderr, "Error %d opening file %s\n", status, filename);
    return status;
  }
  
  if (max_columns!=min_columns) {
    fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
    fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
  }  

  if (min_columns<3) {
    fprintf (stderr, "Error, found less than 3 columns in %s\n",
	     filename);
    return -1;
  }
  

  /* allocate memory for 1D cloud structure */

  if (optical) {
    status = calloc_cld_out (cld, output->wl.nlambda_t, rows);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }
  else {
    status = calloc_cld_microphysics (cld, output->wl.nlambda_t, rows);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }

  for (lev=0; lev<cld->nlev; lev++) {
    cld->zd            [lev] = cld_data[lev][CLD_ZD];
    cld->microphys.lwc [lev] = cld_data[lev][CLD_LWC];
    cld->microphys.effr[lev] = cld_data[lev][CLD_EFFR];
  }
  
  
  /* free temporary memory */
  ASCII_free_float(cld_data, rows);
  
  
  /* Shift clouds to given altitude */
  if (input.atm.altitude != 0) {
    status = shift_cloud (&input.atm, cld);
    if (status !=0) {
      fprintf (stderr, "Error %d shifting clouds\n", status);
      return status;
    }
  }
  
  /* check for consistency with atmospheric levels */
  status = compare_levels (cld->zd, cld->nlev,
			   output->atm.zd, output->atm.nlev);
  
  if (status!=0) {
    fprintf (stderr, "Error: z-grids of atmosphere and 1D clouds are different\n");
    return status;
  }
  
  /* newsiz is a flag, telling the Fortran function wcloud() if a */
  /* new liquid water content and/or effective radius are given   */
  /* (newsiz=1) or if the values haven't changed since the last   */
  /* call (newsiz=0).                                             */ 
  cld->newsiz = 1;
  
  
  return 0;
}
  




/* Read water cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                    */

int read_and_convert_wc_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *wc)
{
  int iv=0, status=0;
  int nstring = (int) strlen(input.filename[FN_PATH]);

  void  F77_FCN (wcloud) (float* lambda_t, int* newsiz, int* nlyr,
			  char* filepath, int* nstring, float* lwc, float* wceffr, 
			  float* tmp_wc_dtau, float* tmp_wc_gg, float* tmp_wc_ssa, float* zd,
			  int *wclyr);

  

  status = read_cld_file (input, output, filename, wc, 1);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_WCLOUD]);
    return status;
  }
  

    
  
  /* calculate cloud optical properties */
  switch (input.ck_scheme) {
  case CK_FU:
    
    /* call Fu and Liou code by Fred Rose */
    
#if HAVE_FULIOU
    status = ckdfucld (wc->microphys.effr, 
		       wc->microphys.lwc, 
		       output->atm.nlev,
		       wc->optprop.dtau, 
		       wc->optprop.gg, 
		       wc->optprop.ssa);
#else
    fprintf (stderr, "Error, Fu and Liou not supported!\n");
    return -1;
#endif
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
      return status;
    }
    
    break;
    
  default:
    
    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      F77_FCN (wcloud) (&output->wl.lambda_t[iv], &(wc->newsiz), &output->atm.nlyr,
			input.filename[FN_PATH], &nstring, 
			wc->microphys.lwc, wc->microphys.effr, 
			wc->optprop.dtau[iv], 
			wc->optprop.gg[iv], wc->optprop.ssa[iv], 
			output->atm.zd, &input.wc.layer);
    }
    
    break;
  }

  return 0;
}
  




/* Read ice cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                  */

int read_and_convert_ic_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *ic)
{
  int iv=0, status=0;

  status = read_cld_file (input, output, filename, ic, 1);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  if (input.ic.properties == CLDPROP_YANG)
    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      
      status = ic_yang (output->wl.lambda_t[iv], output->atm.nlyr,
			input.ic.habit,
			input.filename[FN_PATH],
			ic->microphys.lwc, ic->microphys.effr, 
			ic->optprop.dtau[iv], 
			ic->optprop.f[iv], 
			ic->optprop.g1[iv], ic->optprop.g2[iv], 
			ic->optprop.ssa[iv], 
			output->atm.zd, input.ic.layer);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by ic_yang()\n", status);
	return status;
      }
      
 

    }
  else {
    
    /* calculate cloud optical properties */
    switch (input.ck_scheme) {
    case CK_FU:
      

#if HAVE_FULIOU
      
      status = ckdfuice (ic->microphys.effr, ic->microphys.lwc, output->atm.nlev,
			 ic->optprop.dtau, ic->optprop.gg, ic->optprop.ssa);
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfuice()\n", status);
	return status;
      }
      
      break;
      
    default:
      
      if (strlen(input.filename[FN_ICLOUD_FILES])==0 && strlen(input.filename[FN_ICLOUD_PROP])==0) {
	fprintf (stderr, "Error, need to specify either 'correlated_k FU' or 'ic_files ...'\n");
	fprintf (stderr, "or 'ic_properties default' or or 'ic_properties yang'!\n");
	return -1;
      }
    }
  }

  return 0;
}



/* convert microphysical and single scattering properties */
/* to optical properties at model layers                  */

int ssprop2optprop (cld_out_struct *cld, int iv)
{

  int lc=0, status=0;
  
  
  for (lc=0; lc<cld->nlev-1; lc++) {  /* count layers, 0 ... nlev-2 */
    
    /* first free memory, if already allocated */
    if (cld->optprop.moment[iv][lc] != NULL) {
      free (cld->optprop.moment[iv][lc]);
      cld->optprop.moment[iv][lc] = NULL;
    }

    if (cld->microphys.lwc[lc+1] > 0) {
      
      /* interpolate Legendre coefficients to specified effective radius; */
      /* memory for moments is allocated here in this function.           */
      status = interpolate_legendre (cld->ssprop.legen[iv], 
				     cld->ssprop.nleg[iv], 
				     cld->ssprop.nreff[iv],
				     cld->ssprop.r0[iv], 
				     cld->ssprop.dr[iv],
				     cld->ssprop.extinc[iv], 
				     cld->ssprop.albedo[iv],
				     cld->microphys.effr[lc+1],
				     &(cld->optprop.moment[iv][lc]), 
				     &(cld->optprop.nmom[iv][lc]), 
				     &(cld->optprop.dtau[iv][lc]), 
				     &(cld->optprop.ssa[iv][lc]));
      
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating Legendre coefficients to reff = %f in layer %d.\n", 
		 status, cld->microphys.effr[lc+1], lc);
	return status;
      }
	  

      /* convert from extinction / lwc to optical thickness dtau */
      cld->optprop.dtau[iv][lc] *= cld->microphys.lwc[lc+1] * 
	(cld->zd[lc] - cld->zd[lc+1]);
      
      /* BBMM if (cld->nmom[iv][lc]>0) */
      if (cld->optprop.nmom[iv][lc]>1)
	cld->optprop.gg[iv][lc] = cld->optprop.moment[iv][lc][1];

    }
    else {
      cld->optprop.dtau[iv][lc] = 0;
      cld->optprop.ssa [iv][lc] = 0;
      cld->optprop.gg  [iv][lc] = 0;

      /* reset Legendre coefficients */
      /* BBMM cld->nmom[iv][lc] = 0; */
      /* BBMM cld->moment[iv][lc] = calloc (1, sizeof(double)); */
      /* BBMM cld->moment[iv][lc][0] = 1; */

      cld->optprop.nmom[iv][lc] = 0;
    }
  }
  

  return 0;
}




int read_wcloud_prop (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;

  char momfilename[FILENAME_MAX]="";
  char tempstr[20]="";

  double wavelen=0, nre=0, nim=0;
  
  if (input.verbose) {
    fprintf (stderr, " ... reading water cloud optical properties from %s\n", 
	     input.filename[FN_WCLOUD_PROP]);
  }

  for (iv=0; iv<output->wl.nlambda_t; iv++) {
    
    /* read optical properties from file */
    strcpy (momfilename, input.filename[FN_WCLOUD_PROP]);
    
    if (iv+1<100)
      strcat (momfilename, "0");
    
    if (iv+1<10) 
      strcat (momfilename, "0");
    
    sprintf (tempstr, "%d", iv+1);
    
    strcat (momfilename, tempstr);
    strcat (momfilename, ".mie");
	
    status = read_mie_table (momfilename, 
			     &(output->wc.ssprop.r0[iv]), 
			     &(output->wc.ssprop.dr[iv]), 
			     &(output->wc.ssprop.nreff[iv]),
			     &wavelen, &nre, &nim, 
			     &(output->wc.ssprop.extinc[iv]), 
			     &(output->wc.ssprop.albedo[iv]), 
			     &(output->wc.ssprop.nleg[iv]),
			     &(output->wc.ssprop.legen[iv]), 
			     input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, momfilename);
      return status;
    }
	
    output->wc.ssprop.alloc = 1;
    
    
    /* convert microphysical and single scattering properties */
    /* to optical properties                                  */
    status = ssprop2optprop (&(output->wc), iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating water cloud optical properties\n", status);
      return status;
    }
  }	
  
  /* increase number of moments if more found */
  for (iv=0; iv<output->wl.nlambda_t; iv++)
    for (lc=0; lc<output->atm.nlyr; lc++)
      /* BBMM if (output->wc.nmom[iv][lc] > output->atm.nmom) */
      /* BBMM   output->atm.nmom = output->wc.nmom[iv][lc];     */
      if (output->wc.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	output->atm.nmom = output->wc.optprop.nmom[iv][lc]-1;


  return 0;
}
  






int read_icloud_prop (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;

  char momfilename[FILENAME_MAX]="";
  char tempstr[20]="";

  double wavelen=0, nre=0, nim=0;
  
      

  if (input.verbose) {
    fprintf (stderr, " ... reading ice cloud optical properties from %s\n", 
	     input.filename[FN_ICLOUD_PROP]);
  }

  for (iv=0; iv<output->wl.nlambda_t; iv++) {
    
    /* read optical properties from file */
    strcpy (momfilename, input.filename[FN_ICLOUD_PROP]);
    
    if (iv+1<100)
      strcat (momfilename, "0");
    
    if (iv+1<10) 
      strcat (momfilename, "0");
    
    sprintf (tempstr, "%d", iv+1);
    
    strcat (momfilename, tempstr);
    strcat (momfilename, ".mie");
    
    status = read_mie_table (momfilename, 
			     &(output->ic.ssprop.r0[iv]), 
			     &(output->ic.ssprop.dr[iv]), 
			     &(output->ic.ssprop.nreff[iv]),
			     &wavelen, &nre, &nim, 
			     &(output->ic.ssprop.extinc[iv]), 
			     &(output->ic.ssprop.albedo[iv]), 
			     &(output->ic.ssprop.nleg[iv]),
			     &(output->ic.ssprop.legen[iv]), 
			     input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, momfilename);
      return status;
    }
    
    output->ic.ssprop.alloc = 1;
    
    /* convert microphysical and single scattering properties */
    /* to optical properties                                  */
    status = ssprop2optprop (&(output->ic), iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating ice cloud optical properties\n", status);
      return status;
    }
  }	
  
  /* increase number of moments if more found */
  for (iv=0; iv<output->wl.nlambda_t; iv++)
    for (lc=0; lc<output->atm.nlyr; lc++)
      /* BBMM if (output->ic.nmom[iv][lc] > output->atm.nmom ) */
      /* BBMM output->atm.nmom = output->ic.nmom[iv][lc];     */
      if (output->ic.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	output->atm.nmom = output->ic.optprop.nmom[iv][lc]-1;
  
  return 0;
}





/* overwrite water cloud properties with user-defined properties */
int apply_user_defined_properties_to_cld (cld_inp_struct cldin, 
					  atm_out_struct atmout,
					  wl_out_struct  wlout,
					  cld_out_struct *cldout)
{
  int iv=0, lc=0;
  float scale_factor=0, sum=0;


  for (iv=0; iv<wlout.nlambda_t; iv++) {
    
    /* Set the water cloud optical depth to a constant user-defined value */
    if (cldin.tau >= 0.0) {
      sum = 0.0;
      for (lc=0; lc<atmout.nlyr; lc++)
	sum += cldout->optprop.dtau[iv][lc];
      
      scale_factor = 0.0;
      if (sum>0)
	scale_factor = cldin.tau/sum;
      
      for (lc=0; lc<atmout.nlyr; lc++)
	cldout->optprop.dtau[iv][lc] *= scale_factor;
    }
    
    
    /* Set the water cloud asymmetry factor to a user-defined value */
    if (cldin.gg >= -1.0)
      for (lc=0; lc<atmout.nlyr; lc++)
	cldout->optprop.gg[iv][lc] = cldin.gg;
    
    /* Set the water cloud single scattering albedo to a user-defined value */
    if (cldin.ssa >= 0.0)
      for (lc=0; lc<atmout.nlyr; lc++)
	cldout->optprop.ssa[iv][lc] = cldin.ssa;

    /* Scale the water cloud asymmetry factor with a user-defined value */
    if (cldin.scale_gg >= 0.0)
      for (lc=0; lc<atmout.nlyr; lc++) {
	cldout->optprop.gg[iv][lc] *= cldin.scale_gg;
	if (cldout->optprop.gg[iv][lc]>1.0)
	  cldout->optprop.gg[iv][lc] = 1.0;
      }

    /* Scale the water cloud single scattering albedo with a user-defined value */
    if (cldin.scale_ssa >= 0.0)
      for (lc=0; lc<atmout.nlyr; lc++) {
	cldout->optprop.ssa[iv][lc] *= cldin.scale_ssa;
	if (cldout->optprop.ssa[iv][lc]>1.0)
	  cldout->optprop.ssa[iv][lc] = 1.0;
      }
  }

  return 0;
}


/*agf*/
static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer)
{
  float ext=0;
  int lc=0, status=0;
  char mipath[256]="";
  float maxreff = 55.0;
  
  sprintf(mipath,"%s/ic/yang56/",path);

  for (lc=0; lc<nlyr; lc++) {
  
    /* calculate dtau[lc], g1[lc], g2[lc], f[lc], ssa[lc] from */
    /* iwc[lc], reff[lc], wavelength                           */

    if (iwc[lc]>0.0) {

#if HAVE_YANG56
      switch (habit)  {
      case IC_HABIT_SOLIDCOLUMN:   
	maxreff=MAX_REFF_SC; 
	break;
      case IC_HABIT_HOLLOWCOLUMN:  
	maxreff=MAX_REFF_HC; 
	break;
      case IC_HABIT_AGGREGATE:     
	maxreff=MAX_REFF_RA; 
	break;
      case IC_HABIT_ROSETTE4:      
	maxreff=MAX_REFF_R4; 
	break;
      case IC_HABIT_ROSETTE6:      
	maxreff=MAX_REFF_R6; 
	break;
      case IC_HABIT_PLATE:         
	maxreff=MAX_REFF_PL; 
	break;
      default:
	fprintf (stderr, "Error, unknown habit %d\n", habit);
	return -1;
      }
      
      if (reff[lc]>maxreff) {
	fprintf (stderr, "Error, ice effective radius too large for this habit\n");
	return -1;
      }
      

      status = yang56 (wavelength/1000.0, reff[lc], habit, mipath, &ext, ssa+lc, g1+lc, g2+lc, f+lc);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by yang56()\n", status);
	return status;
      }
#else
      fprintf (stderr, " ***********************************************************************\n");
      fprintf (stderr, " * You have built uvspec without Yang et al support and hence cannot   *\n");
      fprintf (stderr, " * use 'ic_properties yang'.                                           *\n");
      fprintf (stderr, " ***********************************************************************\n");
      return -1;
#endif

      dtau[lc] = ext*iwc[lc]*(zd[lc]-zd[lc+1]);
    }
  }
  
  return 0;   /* if ok */
}





