/*--------------------------------------------------------------------
 * $Id: noon.c,v 1.2 2001/06/20 14:30:58 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "sun.h"


#define IFU_LATITUDE   47.48
#define IFU_LONGITUDE -11.07
#define IFU_LONG_STD  -15.00


/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void print_usage (char *filename)
{
  fprintf (stderr, "NOON 1.99a - calculate local noon time\n\n");
  fprintf (stderr, "written by  Bernhard Mayer,\n");
  fprintf (stderr, "            DLR, eMail bernhard.mayer@dlr.de\n");
  fprintf (stderr, "Version 1.99a finished October 23, 1995\n\n");
  fprintf (stderr, "Be aware that this program is a beta version! Please report any\n");
  fprintf (stderr, "kind of error to the author, including the error message and the\n");
  fprintf (stderr, "database being processed when the error occured. Thanks!\n\n");
  fprintf (stderr, "USAGE: %s [options] <day> <month>\n\n", filename);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "  -h               display help message\n");
  fprintf (stderr, "  -a               Latitude  (North positive)\n");
  fprintf (stderr, "  -o               Longitude (West positive)\n");
  fprintf (stderr, "  -s               Standard Longitude\n\n");
  fprintf (stderr, "If the coordinates are omitted, the location of IFU is used for\n");
  fprintf (stderr, "the calculation:\n");
  fprintf (stderr, "  Latitude 47.48, Longitude -11.07, Standard Longitude.\n");
}



/*************************************************************/
/* parse command line options                                */
/*************************************************************/

static int get_options (int argc, char **argv,
			char *programname,
			int *doy,
			int *print,
			double *latitude,
			double *longitude,
			double *long_std)
{
  int c=0;
  char *dummy=NULL;

  int day   = 0;
  int month = 0;

  char lat=0, lon=0, std=0;


  /* reset parameters to default values */
  *latitude  = IFU_LATITUDE;
  *longitude = IFU_LONGITUDE;
  *long_std  = IFU_LONG_STD;
  

  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);
  
  /* get command line arguments */
  /* The following line is not needed with the getopt.{c,h} I have */
  /* Arve Kylling, Aug 9, 1997 */
  /* optmode = GETOPT_ANY; */

  while ((c=getopt (argc, argv, "ha:o:ps:")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'a':  /* latitude  */  
      *latitude  = (double) strtod (optarg, &dummy);
      lat=1;
      break;   
    case 'o':  /* longitude */  
      *longitude = (double) strtod (optarg, &dummy);
      lon=1;
      break;   
    case 'p':  /* no print messages */  
      *print = 0;
      break;   
    case 's':  /* standard longitude */  
      *long_std = (double) strtod (optarg, &dummy);
      std=1;
      break;   
    case '?':
      print_usage (programname);
      return (-1);
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }
      

  /* check number of remaining command line arguments */
  if (argc - optind != 2)  {
    print_usage (programname);
    return -1;
  }
  
  /* check arguments */
  if (lat&!lon || lat&!std || lon&!lat || lon&!std || std&!lat || std&!lon) {
    print_usage (programname);
    fprintf (stderr, "\n** Specify all three parameters, latitude, longitude and **\n");
    fprintf (stderr,   "** standard longitude!!                                  **\n");
    
    *latitude  = IFU_LATITUDE;
    *longitude = IFU_LONGITUDE;
    *long_std  = IFU_LONG_STD;

    return -1;
  }


  /* check coordinates range */
  if (*latitude<-90 || *latitude>90)  {
    fprintf (stderr, "\n** Latitude out of range **\n");
    return -1;
  }

  if (*longitude<-180 || *longitude>180)  {
    fprintf (stderr, "\n** Longitude out of range **\n");
    return -1;
  }

  if (*long_std<-180 || *long_std>180)  {
    fprintf (stderr, "\n** Standard Longitude out of range **\n");
    return -1;
  }


  /* get command line arguments */
  day    = strtod (argv[optind+0], &dummy);
  month  = strtod (argv[optind+1], &dummy);

  /* calculate day of year */
  if ((*doy = day_of_year(day, month)) < 0)  {
    print_usage (programname);
    fprintf (stderr, "\n** Date out of range **\n");
    return -1;
  }


  return 0;  /* if o.k. */
}






int main(int argc, char ** argv)
{
  char programname[FILENAME_MAX]="";

  double latitude  = 0;
  double longitude = 0;
  double long_std  = 0;
  double zenith    = 0;

  int doy   = 0;
  int hour  = 0;
  int min   = 0;
  int sec   = 0;

  int time=0;
  int shift=0;

  int print =1;
  int status=0;

  char dummy[10] = "";


  /* get command line options */
  status = get_options (argc, argv,
			programname,
			&doy, &print, &latitude, &longitude, &long_std);

  if (status!=0)
    return status;



  /* calculate LOCAL APPARENT TIME for midnight */
  shift = LAT (0, doy, longitude, long_std);
  
  /* calculate local noon time */
  time = 43200 - shift;

  /* calculate solar zenith angle */
  zenith  = solar_zenith  (time, doy, latitude, longitude, long_std);

  hour = time / 3600;
  min  = (time - hour*3600) / 60;
  sec  = (time - hour*3600 - min*60);
    
  if ( print ) {
    fprintf (stderr, "Coordinates: Lat. %8.4f, Long. %8.4f, Standard Long. %8.4f\n",
	     latitude, longitude, long_std);
    
    fprintf (stderr, "  Local Noon Time:       %s CET\n", 
	     time2str(dummy, hour,min,sec));
    fprintf (stderr, "  Solar Zenith Angle:    %f\n\n", zenith);
    fprintf (stderr, " ... printing data to stdout\n");
    fprintf (stderr, " <local noon time>  <local noon time>  <solar zenith angle>\n");
    fprintf (stderr, "    [hh:mm:ss]          [hours]              [degrees]\n");
  }
  fprintf (stdout, "%s  %g  %g\n", time2str (dummy, hour,min,sec), time/3600.0, zenith);


  return 0;
}
